__author__ = 'Robert Meyer'





import numpy as np
import unittest
from pypet.parameter import BaseParameter,Parameter, PickleParameter, BaseResult, ArrayParameter, PickleResult
from pypet.trajectory import Trajectory
from pypet.utils.explore import cartesian_product
from pypet.environment import Environment
from pypet import globally
import logging
from pypet.utils.helpful_functions import flatten_dictionary
from pypet.utils.comparisons import results_equal,parameters_equal
import scipy.sparse as spsp
import os
import shutil
import pandas as pd
import tables as pt


REMOVE = True
SINGLETEST=[0]


def simple_calculations(traj, arg1, simple_kwarg):


        # all_mat = traj.csc_mat + traj.lil_mat + traj.csr_mat
        Normal_int= traj.Normal.int
        Sum= np.sum(traj.Numpy.double)

        # result_mat = all_mat * Normal_int * Sum * arg1 * simple_kwarg



        my_dict = {}

        my_dict2={}
        for key, val in traj.parameters.f_to_dict(fast_access=True,short_names=False).items():
            if 'trial' in key:
                continue
            newkey = key.replace('.','_')
            my_dict[newkey] = str(val)
            my_dict2[newkey] = [str(val)+' juhu!']

        my_dict['__FLOAT'] = 44.0
        my_dict['__INT'] = 66
        my_dict['__NPINT'] = np.int_(55)
        my_dict['__INTaRRAy'] = np.array([1,2,3])
        my_dict['__FLOATaRRAy'] = np.array([1.0,2.0,41.0])
        my_dict['__STRaRRAy'] = np.array(['sds','aea','sf'])

        keys = traj.f_to_dict(short_names=False).keys()
        for idx,key in enumerate(keys):
            keys[idx] = key.replace('.','_')

        traj.f_add_result('List.Of.Keys', dict1=my_dict, dict2=my_dict2)
        traj.f_add_result('DictsNFrame', keys=keys)
        traj.f_add_result('ResMatrix',np.array([1.2,2.3]))
        #traj.f_add_derived_parameter('All.To.String', str(traj.f_to_dict(fast_access=True,short_names=False)))

        myframe = pd.DataFrame(data ={'TC1':[1,2,3],'TC2':['Waaa',np.nan,''],'TC3':[1.2,42.2,np.nan]})

        traj.DictsNFrame.f_set(myframe)

        traj.f_add_result('IStore.SimpleThings',1.0,3,np.float32(5.0), 'Iamstring',(1,2,3),[4,5,6],zwei=2)
        traj.f_add_derived_parameter('mega',33)

        #traj.f_add_result('PickleTerror', result_type=PickleResult, test=traj.SimpleThings)


class ContinueTest(unittest.TestCase):



    def compare_trajectories(self,traj1,traj2):

        old_items = traj1.f_to_dict(fast_access=False)
        new_items = traj2.f_to_dict(fast_access=False)



        self.assertEqual(len(old_items),len(new_items))
        for key,item in new_items.items():
            old_item = old_items[key]
            if key.startswith('config'):
                continue

            if isinstance(item, BaseParameter):
                self.assertTrue(parameters_equal(item,old_item),
                                'For key %s: %s not equal to %s' %(key,str(old_item),str(item)))
            elif isinstance(item,BaseResult):
                self.assertTrue(results_equal(item, old_item),
                                'For key %s: %s not equal to %s' %(key,str(old_item),str(item)))
            else:
                raise RuntimeError('You shall not pass')

    def make_run(self,env):

        ### Make a test run
        simple_arg = -13
        simple_kwarg= 13.0
        env.f_run(simple_calculations,simple_arg,simple_kwarg=simple_kwarg)

    def _create_param_dict(self):
        self.param_dict = {}

        self.param_dict['Normal'] = {}
        self.param_dict['Numpy'] = {}
        # self.param_dict['Sparse'] ={}
        self.param_dict['Numpy_2D'] = {}
        self.param_dict['Numpy_3D'] = {}
        self.param_dict['Tuples'] ={}

        normal_dict = self.param_dict['Normal']
        normal_dict['string'] = 'Im a test string!'
        normal_dict['int'] = 42
        normal_dict['double'] = 42.42
        normal_dict['bool'] =True
        normal_dict['trial'] = 0

        numpy_dict=self.param_dict['Numpy']
        numpy_dict['string'] = np.array(['Uno', 'Dos', 'Tres'])
        numpy_dict['int'] = np.array([1,2,3,4])
        numpy_dict['double'] = np.array([1.0,2.0,3.0,4.0])
        numpy_dict['bool'] = np.array([True,False, True])

        #self.param_dict['Numpy_2D']['double'] = np.array([[1.0,2.0],[3.0,4.0]])
        #self.param_dict['Numpy_3D']['double'] = np.array([[[1.0,2.0],[3.0,4.0]],[[3.0,-3.0],[42.0,41.0]]])

        # spsparse_csc = spsp.csc_matrix((2222,22))
        # spsparse_csc[1,2] = 44.6
        #
        # spsparse_csr = spsp.csr_matrix((2222,22))
        # spsparse_csr[1,3] = 44.7
        #
        # spsparse_lil = spsp.lil_matrix((2222,22))
        # spsparse_lil[3,2] = 44.5
        #
        # self.param_dict['Sparse']['lil_mat'] = spsparse_lil
        # self.param_dict['Sparse']['csc_mat'] = spsparse_csc
        # self.param_dict['Sparse']['csr_mat'] = spsparse_csr

        self.param_dict['Tuples']['int'] = (1,2,3)
        self.param_dict['Tuples']['float'] = (44.4,42.1,3.)
        self.param_dict['Tuples']['str'] = ('1','2wei','dr3i')


    def add_params(self,traj):

        flat_dict = flatten_dictionary(self.param_dict,'.')

        for key, val in flat_dict.items():
            if isinstance(val, (np.ndarray,list, tuple)):
                traj.f_add_parameter(ArrayParameter,key,val)
            elif isinstance(val, (int,str,bool,float)):
                traj.f_add_parameter(Parameter,key,val, comment='Im a comment!')
            elif spsp.isspmatrix(val):
                traj.f_add_parameter(PickleParameter,key,val)
            else:
                raise RuntimeError('You shall not pass, %s is %s!' % (str(val),str(type(val))))

        traj.f_add_derived_parameter('Another.String', 'Hi, how are you?')
        traj.f_add_result('Peter_Jackson',np.str(['is','full','of','suboptimal ideas']),comment='Only my opinion bro!',)

    def make_environment(self, idx, filename):

        logging.basicConfig(level = logging.DEBUG)

        #self.filename = '../../experiments/tests/HDF5/test.hdf5'
        logfolder = '../../experiments/tests/Log'
        trajname = 'Test%d' % idx

        env = Environment(trajectory=trajname,filename=filename,file_title=trajname, log_folder=logfolder)


        self.envs.append(env)
        self.trajs.append( env.v_trajectory)


    def explore(self, traj):
        self.explored ={'Normal.trial': [0,1],
            'Numpy.double': [np.array([1.0,2.0,3.0,4.0]), np.array([-1.0,3.0,5.0,7.0])]}


        traj.f_explore(cartesian_product(self.explored))

    @unittest.skipIf(SINGLETEST != [0] and 1 not in SINGLETEST,'Skipping because, '
                                                               'single debug is not pointing to the function ')
    def test_continueing(self):
        self.filenames = ['../../experiments/tests/HDF5/merge1.hdf5', 0]



        self.envs=[]
        self.trajs = []

        for irun,filename in enumerate(self.filenames):
            if isinstance(filename,int):
                filename = self.filenames[filename]

            self.make_environment( irun, filename)

        self._create_param_dict()

        for irun in range(len(self.filenames)):
            self.add_params(self.trajs[irun])


        self.explore(self.trajs[0])
        self.explore(self.trajs[1])

        for irun in range(len(self.filenames)):
            self.make_run(self.envs[irun])


        ### Create a crash and say, that the first and last run did not work.
        pt_file = pt.open_file(self.filenames[0],mode='a')
        runtable = pt_file.get_node('/'+self.trajs[0].v_name+'/run_table')

        for idx,row in enumerate(runtable.iterrows()):
            if idx == 0 or idx == 3:
                row['completed'] = 0
                row.update()

        runtable.flush()
        pt_file.flush()
        pt_file.close()


        continue_file = os.path.split(self.filenames[0])[0]+'/'+self.trajs[0].v_name+'.cnt'
        self.envs[0].f_continue_run(continue_file)

        for irun in range(len(self.filenames)):
            self.trajs[irun].f_update_skeleton()
            self.trajs[irun].f_load(load_parameters=globally.UPDATE_DATA,
                                    load_derived_parameters=globally.UPDATE_DATA,
                                    load_results=globally.UPDATE_DATA)

        self.compare_trajectories(self.trajs[0],self.trajs[1])

    @unittest.skipIf(SINGLETEST != [0] and 2 not in SINGLETEST,'Skipping because, '
                                                               'single debug is not pointing to the function ')
    def test_removal(self):
        self.filenames = ['../../experiments/tests/HDF5/merge1.hdf5', 0]



        self.envs=[]
        self.trajs = []

        for irun,filename in enumerate(self.filenames):
            if isinstance(filename,int):
                filename = self.filenames[filename]

            self.make_environment( irun, filename)

        self._create_param_dict()

        for irun in range(len(self.filenames)):
            self.add_params(self.trajs[irun])


        self.explore(self.trajs[0])
        self.explore(self.trajs[1])





        for irun in range(len(self.filenames)):
            self.make_run(self.envs[irun])

        self.trajs[0].f_add_parameter('Delete.Me', 'I will be deleted!')
        self.trajs[0].f_store_item('Delete.Me')

        self.trajs[0].f_remove_item(self.trajs[0].f_get('Delete.Me'),
                                        remove_empty_groups=True)

        self.assertTrue(not 'Delete.Me' in self.trajs[0],'Delete.Me is still in traj')

        self.trajs[0].f_update_skeleton()
        self.trajs[0].f_load_item('Delete.Me')
        self.trajs[0].f_remove_item(self.trajs[0].f_get('Delete.Me'), remove_from_storage=True,
                                        remove_empty_groups=True)

        self.trajs[0].f_update_skeleton()
        self.assertTrue(not 'Delete.Me' in self.trajs[0],'Delete.Me is still in traj')


        for irun in range(len(self.filenames)):
            self.trajs[irun].f_update_skeleton()
            self.trajs[irun].f_load_child('results',recursive=True,load_data=globally.UPDATE_DATA)
            self.trajs[irun].f_load_child('derived_parameters',recursive=True,load_data=globally.UPDATE_DATA)

        self.compare_trajectories(self.trajs[0],self.trajs[1])

    @unittest.skipIf(SINGLETEST != [0] and 3 not in SINGLETEST,'Skipping because, '
                                                'single debug is not pointing to the function ')
    def test_multiple_storage_and_loading(self):
        self.filenames = ['../../experiments/tests/HDF5/merge1.hdf5', 0]



        self.envs=[]
        self.trajs = []

        for irun,filename in enumerate(self.filenames):
            if isinstance(filename,int):
                filename = self.filenames[filename]

            self.make_environment( irun, filename)

        self._create_param_dict()

        for irun in range(len(self.filenames)):
            self.add_params(self.trajs[irun])


        self.explore(self.trajs[0])
        self.explore(self.trajs[1])





        for irun in range(len(self.filenames)):
            self.make_run(self.envs[irun])

        #self.trajs[0].f_store()

        temp_sservice = self.trajs[0].v_storage_service
        temp_name = self.trajs[0].v_name

        self.trajs[0] = Trajectory()
        self.trajs[0].v_storage_service=temp_sservice
        self.trajs[0].f_load(trajectory_name=temp_name,as_new=False, load_parameters=2, load_derived_parameters=2, load_results=2)
        #self.trajs[0].f_load(trajectory_name=temp_name,as_new=False, load_params=2, load_derived_params=2, load_results=2)

        self.trajs[1].f_update_skeleton()
        self.trajs[1].f_load_items(self.trajs[1].f_to_dict().itervalues(),only_empties=True)
        self.compare_trajectories(self.trajs[0],self.trajs[1])


if __name__ == '__main__':
    if REMOVE:
        shutil.rmtree('../../Test',True)
    unittest.main()


'''
Created on 17.05.2013

@author: robert
'''

import logging
import datetime
import time
from pypet.parameter import Parameter, BaseParameter, Result, BaseResult, ArrayParameter, \
    PickleResult

import importlib as imp
import itertools as it
import inspect
import numpy as np

import pypet.petexceptions as pex
from pypet import globally
from pypet.naturalnaming import NNGroupNode,NaturalNamingInterface, ResultGroup, ParameterGroup, \
    DerivedParameterGroup, ConfigGroup, FORMATTED_RUN_NAME,STORE,LOAD,REMOVE

from pypet.storageservice import HDF5StorageService





class SingleRun(DerivedParameterGroup,ResultGroup):
    ''' Constitutes one specific parameter combination in the whole trajectory.

    A SingleRun instance is accessed during the actual run phase of a trajectory.
    There exists a SignleRun object for each point in the parameter space.
    The object contains a copy of the trajectory. How large that copy is depends on
    `v_full_copy` in the parent trajectory. If `v_full_copy=True` than the full
    trajectory can be accessed. Otherwise from each parameter array
    only the idxth parameter can be accesses, where idx is the index of the current run.

    parameters can no longer be added, the parameter set is supposed to be complete before
    a the actual running of the experiment. However, derived parameters can still be added.
    This might be useful, for example, to store a connectivity matrix of neural network,
    that is built new for each point in the trajectory.


    The instance of a SingleRun is never instantiated by the user but by the parent trajectory.

    '''
    def __init__(self, name, idx, parent_trajectory):


        super(SingleRun,self).__init__(root=True)

        self._parent_trajectory = parent_trajectory

        #if init_time is None:
        init_time = time.time()

        formatted_time = datetime.datetime.fromtimestamp(init_time).strftime('%Y_%m_%d_%Hh%Mm%Ss')

        self._name = name
        self._timestamp = init_time
        self._time = formatted_time
        self._idx = idx

        self._trajectory_name =parent_trajectory.v_name
        self._trajectory_time = parent_trajectory.v_time
        self._trajectory_timestamp = parent_trajectory.v_timestamp


        self._parameters = parent_trajectory._parameters
        self._derived_parameters = parent_trajectory._derived_parameters
        self._results = parent_trajectory._results
        self._explored_parameters = parent_trajectory._explored_parameters
        self._config = parent_trajectory._config
        self._groups=parent_trajectory._groups



        self._nn_interface=parent_trajectory._nn_interface
        self._nn_interface._change_root(self)


        self._changed_default_parameters={}

        self._storage_service = parent_trajectory._storage_service

        self._standard_parameter = parent_trajectory.v_standard_parameter
        self._standard_result = parent_trajectory.v_standard_result
        self._search_strategy = parent_trajectory.v_search_strategy
        self._check_uniqueness = parent_trajectory.v_check_uniqueness
        self._fast_access = parent_trajectory.v_fast_access

        self._stored = False

        self._dynamic_imports = parent_trajectory._dynamic_imports

        self._logger = logging.getLogger('pypet.trajectory.SingleRun=' + self.v_name)

        self._is_run = True
        self._annotations = None


    def __getstate__(self):
        result = self.__dict__.copy()
        del result['_logger']
        return result

    def __setstate__(self, statedict):
        self.__dict__.update(statedict)
        self._logger = logging.getLogger(
            'pypet.trajectory.SingleRun=' + self.v_name)


    def __len__(self):
        ''' Length of a single run can only be 1 and nothing else!
        :return:
        '''
        return 1

    @property
    def v_stored(self):
        '''Whether or not the trajectory or run has been stored to disk before.'''
        return self._stored

    @property
    def v_search_strategy(self):
        '''Search strategy for lookup of items in the trajectory tree.

        Default is breadth first search (BFS), you could alos chose depth first search (DFS),
        but not recommended!

        '''
        return self._search_strategy

    @v_search_strategy.setter
    def v_search_strategy(self,strategy):
        if not strategy == globally.BFS or strategy == globally.DFS:
            raise ValueError('Please use strategies %s or %s others are not supported atm.' %
                             (globally.BFS,globally.DFS))

        self._search_strategy = strategy


    @property
    def v_timestamp(self):
        '''Float timestamp of creation time'''
        return self._timestamp

    @property
    def v_time(self):
        '''Formatted time string of the time the trajectory or run was created.
        '''
        return self._time

    @property
    def v_standard_parameter(self):
        ''' The standard parameter used for parameter creation'''
        return self._standard_parameter

    @v_standard_parameter.setter
    def v_standard_parameter(self, parameter):
        self._standard_parameter = parameter

    @property
    def v_standard_result(self):
        ''' The standard result class used for result creation '''
        return self._standard_result

    @v_standard_result.setter
    def v_standard_result(self, result):
        self._standard_result = result

    @property
    def v_fast_access(self):
        '''Whether parameter instances (False) or their values (True) are returned via natural naming

        Default is True.
        '''
        return self._fast_access

    @v_fast_access.setter
    def v_fast_access(self, value):
        self._fast_access=bool(value)

    @property
    def v_check_uniqueness(self):
        '''Whether natural naming should check if naming is unambiguous

        Default is False. If True, searching a parameter or result via
        :func'~pypet.naturalnaming.NNGroupNode.f_get` will take O(N), because all nodes have
        to be visited!

        '''
        return self._check_uniqueness

    @v_check_uniqueness.setter
    def v_check_uniqueness(self, val):
        self._check_uniqueness = bool(val)


    @property
    def v_storage_service(self):
        return self._storage_service


    @staticmethod
    def _return_item_dictionary(param_dict,fast_access,copy):

        if not copy and fast_access:
            raise TypeError('You cannot access the original dictionary and use fast access at the'
                            'same time!')
        if not fast_access:
            if copy:
                return param_dict.copy()
            else:
                return param_dict
        else:
            resdict = {}
            for key, param in param_dict.iteritems():
                val = param.f_get()
                resdict[key] = val

            return resdict



    @property
    def v_trajectory_name(self):
        '''Name of the parent trajectory'''
        return self._trajectory_name

    @property
    def v_trajectory_time(self):
        ''' Time trajectory was created
        '''
        return self._trajectory_time

    @property
    def v_trajectory_timestamp(self):
        ''' Timestamp when trajectory was created
        '''
        return self._trajectory_timestamp

    @property
    def v_idx(self):
        '''Index of the single run'''
        return self._idx

    def _finalize(self):

        if 'results.'+self.v_name in self:
            self.results.f_remove_child(self.v_name,recursive=True)

        if 'derived_parameters.' +self.v_name in self:
            self.derived_parameters.f_remove_child(self.v_name,recursive=True)


    def f_to_dict(self,fast_access = False, short_names=False, copy = True):
        ''' Returns a dictionary with pairings of (full) names as keys and instances/values.

        :param fast_access: If True, parameter values are returned instead of the instances-

        :param short_names: If true, keys are not full names but only the names. Raises a Value
                            Error if the names are not unique.

        :param copy: If `fast_access=False` and `short_names=False` you can access the original
                    data dictionary if you set `copy=False`. If you do that, please do not
                    modify anything! Raises Value Error if `copy=False` and `fast_access=True`
                    or `short_names=True`.



        :return: dictionary

        :raises: ValueError

        '''
        return self._nn_interface._to_dict(self, fast_access=fast_access,short_names=short_names)

    def f_get_config(self, fast_access=False, copy=True):
        ''' Returns a dictionary containing the full config names as keys and the config parameters
         or the config parameter values.

        :param fast_access: Determines whether the parameter objects or their values are returned
         in the dictionary.

        :param copy: Whether the original dictionary or a shallow copy is returned.
                     If you want the real dictionary please do not modify it at all!
                     Not copy and fast access at the same time do not work! Raises TypeError
                     if fast access is true and copy false.

        :return: Dictionary containing the config data

        :raises: TypeError

        '''
        return self._return_item_dictionary(self._config,fast_access, copy)

    def f_get_parameters(self, fast_access=False,copy=True):
        ''' Returns a dictionary containing the full parameter names as keys and the parameters
         or the parameter values.

         Requires O(n)

        :param fast_access: Determines whether the parameter objects or their values are returned
         in the dictionary.

        :param copy: Whether the original dictionary or a shallow copy is returned.
                     If you want the real dictionary please do not modify it at all!
                     Not using copy and fast access at the same time does not work! Raises TypeError
                     if fast access is true and copy false.

        :return: Dictionary containing the parameters.

        :raises: TypeError

        '''
        return self._return_item_dictionary(self._parameters,fast_access)


    def f_get_explored_parameters(self, fast_access=False, copy=True):
        ''' Returns a dictionary containing the full parameter names as keys and the parameters
         or the parameter values.

         Requires O(n)

        :param fast_access: Determines whether the parameter objects or their values are returned
         in the dictionary.

        :param copy: Whether the original dictionary or a shallow copy is returned.
                     If you want the real dictionary please do not modify it at all!
                     Not using copy and fast access at the same time does not work! Raises TypeError
                     if fast access is true and copy false.

        :return: Dictionary containing the parameters.

        :raises: TypeError

        '''
        return self._return_item_dictionary(self._explored_parameters,fast_access,copy)

    def f_get_derived_parameters(self, fast_access=False, copy=True):
        ''' Returns a dictionary containing the full parameter names as keys and the parameters
         or the parameter values.

         Requires O(n)

        :param fast_access: Determines whether the parameter objects or their values are returned
         in the dictionary.

        :param copy: Whether the original dictionary or a shallow copy is returned.
                     If you want the real dictionary please do not modify it at all!
                     Not using copy and fast access at the same time does not work! Raises TypeError
                     if fast access is true and copy false.

        :return: Dictionary containing the parameters.

        :raises: TypeError

        '''
        return self._return_item_dictionary(self._derived_parameters, fast_access, copy)

    def f_get_results(self,copy=True):
        ''' Returns a dictionary containing the full result names as keys and the corresponding
        result objects.

        :param copy: Whether the original dictionary or a shallow copy is returned.
                     If you want the real dictionary please do not modify it at all!


        :return: Dictionary containing the results.
        '''
        return self._return_item_dictionary(self._results, False,copy)



    def f_store(self, *args, **kwargs):
        ''' Stores the single run to disk
        '''
        #self._srn_add_explored_params()
        self._storage_service.store(globally.SINGLE_RUN, self, trajectory_name=self.v_trajectory_name)

    def f_store_item(self,item,*args,**kwargs):
        ''' Stores a single item, see also :func:`~pypet.trajectory.SingleRun.f_store_items`.
        '''
        self.f_store_items([item],*args,**kwargs)

    def f_store_items(self, iterator, *args, **kwargs):
        ''' Stores individual items to disk.

        This function is useful is you calculated very large results (or large derived parameters)
        during runtime and you want to write these to disk immediately and empty them afterwards
        to free some memory.


        :param iterator: an iterable containing the parameters or results to store, either their
                        names or the instances.

        :param non_empties: Will only store the subset of provided items that are not empty.

        :param args: additional arguments passed to the storage service

        :param kwargs: additional keyword arguments passed to the storage service

        :raises: TypeError: If the (parent) trajectory has never been stored to disk. In this case
                            use :func:'pypet.trajectory.f_store` first.

                ValueError: If no item could be found to be stored.

        Note if you use the standard hdf5 storage service, there are no additional arguments
        or keyword arguments to pass!

        '''
        if self._is_run:
            if not self._stored:
                raise TypeError('Cannot f_store stuff for a trajectory that has never been '
                                'stored to disk. Please call traj.f_store() first, which will '
                                'actually cause the storage of all items in the trajectory.')


        # if iterator == globally.ALL:
        #     iterator = self._all_current.itervalues()

        fetched_items = self._nn_interface._fetch_items(STORE, iterator, args, kwargs)

        if fetched_items:
            self._storage_service.store(globally.LIST, fetched_items, trajectory_name=self.v_trajectory_name)
        else:
            raise ValueError('Your storage was not successful, could not find a single item '
                                 'to store.')




    def _create_class(self, class_name):
        ''' Dynamically creates a class.

        It is tried if the class can be created by default, if not the list of the dynamically
        loaded classes is used (see __init__).
        '''
        try:
            new_class = eval(class_name)
            return new_class
        except NameError:
            for dynamic_class in self._dynamic_imports:
                if inspect.isclass(dynamic_class):
                    if class_name == dynamic_class.__name__:
                        return dynamic_class
                else:
                    class_name_to_test = dynamic_class.split('.')[-1]
                    if class_name == class_name_to_test:
                        new_class = self._load_class(dynamic_class)
                        return new_class
            raise ImportError('Could not create the class named >>%s<<.' % class_name)


class Trajectory(SingleRun,ParameterGroup,ConfigGroup):
    '''The trajectory manages results and parameters.


    The trajectory is the common object to interact with before and during a simulation.
    You can add four types of data to the trajectory:

    * Config:

        These are special parameters specifying modalities of how to run your simulations.
        Changing a config parameter should NOT have any influence on the results you
        obtain from your simulations.

        They specify runtime environment parameters like how many CPUs you use for
        multiprocessing, where the storage file into on disk can be found, etc.

        In fact, if you use the default runtime environment of this project, the environment
        will add some config parameters to your trajectory.

        The method to add these is :func:`~pypet.naturalnaming.ConfigGroup.f_add_config`

    * Parameters:

        These are your primary ammunition in numerical simulations. They specify
        how your simulation works. They can only be added before the actual
        running of the simulation exploring the parameter space. They can be
        added via :func:`~pypet.naturalnaming.ParameterGroup.f_add_parameter`
        and be explored using :func:`~pypet.trajectory.Trajectory.f_explore`.
        Or to expand an existing trajectory use :func:`~pypet.trajectory.Trajectory.f_expand`.

        Your parameters should encompass all values that completely define your simulation,
        I recommend also storing random number generator seeds as parameters to
        guarantee that a simulation can be exactly repeated.



    * Derived Parameters:

        They are not much different from parameters except that they can be
        added anytime.

        Conceptually this encompasses stuff that is intermediately
        computed from the original parameters. For instance, as your original
        parameters you have a random number seed and some other parameters.
        From these you compute a connection matrix for a neural network.
        This connection matrix could be stored as a derived parameter.

        Derived parameters are added via
        :func:`~pypet.naturalnaming.DerivedParameterGroup.f_add_derived_parameter`

    * Results:

        Result are added via the :func:`~pypet.naturalnaming.ResultGroup.f_add_result`

    There are several ways to access the parameters, to learn about these, fast access, and natural
    naming see :ref:`more-on-access`


    
    :param name:

        Name of the trajectory, if `add_time=True` the current time is added as a string
        to the parameter name.


    :param add_time:

        Boolean whether to add the current time to the trajectory name.

    :param comment:

        A useful comment describing the trajectory.

    :param dynamicly_imported_classes:

        If the user has a custom parameter that needs to be loaded
        dynamically during runtime, the module containing the class
        needs to be specified here as a list of classes or strings
        naming classes and there module paths.
        For example:
        `dynamically_imported_classes =
        ['pypet.parameter.PickleParameter',MyCustomParameter]`

        If you only have a single class to import, you do not need
        the list brackets:
        `dynamically_imported_classes
        = 'pypet.parameter.PickleParameter'`






    :param filename:

        If you want to use the default :class:`HDF5StorageService`, you can specify the
        filename of the HDF5 file. If you specify the file name, the trajectory
        will automatically create the corresponding service object.

    :param file_title:

        HDF5 also let's you specify the title of the file.

    :raises:

        AttributeError: If the name of the trajectory contains invalid characters.

        TypeError: If the dynamically imported classes are not classes or strings.
    
                                      
    Example usage:

    >>> traj = Trajectory('ExampleTrajectory',dynamically_imported_classes=['Some.custom.class'],\
    comment = 'I am a neat example!', filename='experiment.hdf5', file_title='Experiments')

    '''
    def __init__(self, name='my_trajectory',add_time=True,comment='', dynamically_imported_classes=None,
        filename=None, file_title = None):


        super(SingleRun,self).__init__(root=True)
        #if init_time is None:
        init_time = time.time()

        formatted_time = datetime.datetime.fromtimestamp(init_time).strftime('%Y_%m_%d_%Hh%Mm%Ss')

        self._timestamp = init_time
        self._time = formatted_time



        if add_time:
            self._name = name + '_' + str(formatted_time)
        else:
            self._name = name

        self._trajectory_name =self.v_name
        self._trajectory_time = self.v_time
        self._trajectory_timestamp = self.v_timestamp


        self._parameters = {}
        self._derived_parameters = {}
        self._results = {}
        self._explored_parameters = {}
        self._config = {}
        self._groups={}


        self._changed_default_parameters = {}

        self._single_run_ids = {}
        self._run_information = {}

        # Per Definition the lenght is set to be 1, even with an _empty trajectory in principle you could make a single run
        self._add_run_info(0)

        self._nn_interface = NaturalNamingInterface(root_instance=self)
        self._fast_access=True
        self._check_uniqueness=False
        self._search_strategy=globally.BFS

        if filename is None:
            self._storage_service=None
        else:
            if file_title is None:
                file_title = filename
            self._storage_service = HDF5StorageService(filename=filename,file_title=file_title)



        self._idx = -1

        self._standard_parameter = Parameter
        self._standard_result = Result

        self._stored = False
        self._full_copy = False

        self._dynamic_imports = ['pypet.parameter.PickleParameter']

        self._is_run = False

        self._length_during_run = None

        if not dynamically_imported_classes is None:
            self.f_add_to_dynamic_imports(dynamically_imported_classes)


        # self._loadedfrom = 'None'


        faulty_names = self._nn_interface._check_names([name])

        if '.' in name:
            faulty_names += ' colons >>.<< are  not allowed in trajectory names,'

        if faulty_names:
            raise AttributeError('Your Trajectory %s f_contains the following not admissible names: '
                                 '%s please choose other names.'
                                 % (name, faulty_names))

        self._logger = logging.getLogger('pypet.trajectory.Trajectory=' + self.v_name)

        self._comment=''
        self.v_comment=comment

        self.f_add_parameter_group('parameters')
        self.f_add_config_group('config')
        self.f_add_result_group('results')
        self.f_add_derived_parameter_group('derived_parameters')


    @property
    def v_comment(self):
        ''' Should be a nice descriptive comment'''
        return self._comment

    @v_comment.setter
    def v_comment(self, comment):

        comment = str(comment)
        if len(comment)>=globally.HDF5_STRCOL_MAX_COMMENT_LENGTH:
            raise AttributeError('Your comment is too long ( %d characters), only comments up to'
                                 '%d characters are allowed.' %
                                 (len(comment),globally.HDF5_STRCOL_MAX_COMMENT_LENGTH))

        self._comment=comment


    @property
    def v_storage_service(self):
        '''The service that can store the trajectory to disk or wherever.

        Default is the :class:`~pypet.storageservice.HDF5StorageService`.
        '''
        return self._storage_service

    @v_storage_service.setter
    def v_storage_service(self,service):
        self._storage_service = service





    @property
    def v_full_copy(self):
        '''Whether trajectory is copied fully during pickling or only the current parameter space point.'''
        return self._full_copy

    @v_full_copy.setter
    def v_full_copy(self, val):
        ''' Sets full copy mode of trajectory and (!) ALL explored parameters!'''
        if val != self._full_copy:
            self._full_copy = val
            for param in self._explored_parameters.itervalues():
                param.v_full_copy = val


    def f_add_to_dynamic_imports(self, dynamically_imported_classes):
        ''' Adds classes or paths to classes to the trajectory to create custom parameters.

        :param dynamically_imported_classes:

            If the user has a custom parameter or result
            that needs to be loaded
            dynamically during runtime, the module containing the class
            needs to be specified here as a list of classes or strings
            naming classes and there module paths.
            For example:
            `dynamically_imported_classes =
            ['pypet.parameter.PickleParameter',MyCustomParameter]`

            If you only have a single class to import, you do not need
            the list brackets:
            dynamically_imported_classes
            = 'pypet.parameter.PickleParameter'`

        '''

        if not isinstance(dynamically_imported_classes, (list, tuple)):
            dynamically_imported_classes = [dynamically_imported_classes]

        for item in dynamically_imported_classes:
            if not (isinstance(item, basestring) or inspect.isclass(item)):
                raise TypeError('Your dynamic import >>%s<< is neither a class nor a string.' %
                                str(item))

        self._dynamic_imports.extend(dynamically_imported_classes)


    # def __iter__(self):
    #     ''' Iterator over all single runs.
    #
    #     equivalent to calling :func:`~pypet.trajectory.Trajectory.f_iter_runs`:
    #
    #         >>> traj.f_iter_runs(non_completed=False)
    #
    #     '''
    #     return self.f_iter_runs(non_completed=False)

    # def f_finalize(self):
    #     ''' If you called :func:` ~pypet.trajectory.Trajectory.f_iter_runs` without completing a full
    #     iteration you are advised to call this function.
    #
    #     Resets all parameters to the default values and restores proper natural naming.
    #     '''
    #     self._finalize()
    #
    # def f_iter_runs(self, non_completed=False):
    #     ''' Returns an Iterator over all single runs.
    #
    #     Will also lock all parameters and reset them to the default values after
    #     the iteration is completed.
    #
    #     Note that this manipulates your original trajectory!
    #     If you break the iteration and want to restore your original trajectory call
    #     :func:`~pypet.trajectory.Trajectory.f_finalize()
    #
    #     :param non_completed: Whether only not completed runs should be considered or not.
    #
    #     '''
    #     self.f_lock_parameters()
    #     self.f_lock_derived_parameters()
    #     for idx in xrange(len(self)):
    #         if (not non_completed) or self.f_get_run_information(idx)['completed'] == 0:
    #             yield self._make_single_run(idx)
    #
    #     self._finalize()



    def f_idx_to_run(self, name_or_idx):
        ''' Converts an integer idx to the corresponding single run name and vice versa.

        :param name_or_idx: Name of a single run of an integer index

        :return: The corresponding idx or name of the single run

        '''
        return self._single_run_ids[name_or_idx]


    def f_get_run_names(self, sort=True):
        '''Returns a list of run names.

        :param sort:

            Whether to get them sorted, will only require O(N) [and not O(N*log N)] since we
            use (sort of) bucket sort.
            Yet, will still be slower than
            `sort=False` because list comprehension is used

        '''
        if sort:
            return [self.f_idx_to_run(idx) for idx in xrange(len(self))]
        else:
            return self._run_information.keys()


    def f_get_run_information(self, name_or_idx=None, copy=True):
        ''' Returns a dictionary containing information about a single run.

        The information dictionaries have the following key, value pairings:

            * completed: Boolean, whether a run was completed.

            * idx: Index of a run

            * timestamp: Timestamp of the run as a float.

            * time: Formatted time string

            * name: Name of the run

            * parameter_summary: A string summary of the explored parameter settings for the particular run.


        If no name or idx is given than a list of all dictionaries is returned.

        :param name_or_idx: str or int

        :param copy:

            Whether you want the dictionary used by the trajectory or a copy. Note if
            you want the real thing, please do not modify it, i.e. popping or adding stuff. This
            could mess up your whole trajectory.

        :return:

            A run information dictionary or a nested dictionary of information dictionaries
            with the run names as keys.

        '''
        if name_or_idx is None:
            if copy:
                return copy.deepcopy(self._run_information)
            else:
                return self._run_information
        try:
            if copy:
                return self._run_information[name_or_idx].copy()
            else:
                return self._run_information[name_or_idx]
        except KeyError:
            name_or_idx = self.f_idx_to_run(name_or_idx)
            if copy:
                return self._run_information[name_or_idx].copy()
            else:
                return self._run_information[name_or_idx]


    def f_remove_item(self, item,*args,**kwargs):
        '''Removes a single item, see :func:`remove_items`'''
        self.f_remove_items([item],*args,**kwargs)

    def f_remove_items(self, iterable, *args, **kwargs):
        '''Removes parameters, results or groups from the trajectory.

        This function can also be used to erase data from disk via the storage service.

        :param iterable: A sequence of items you want to remove. Either the instances themselves
                         or strings with the names of the item.

        :param remove_from_storage: Boolean whether you want to also delete the item from your
                                    storage.
        :param remove_empty_groups: If your deletion of the instance leads to empty groups,
                                    these will be deleted, too.

        :param args:    Additional arguments passed to the storage service

        :param kwargs:  Additional keyword arguments passed to the storage service

        Note if you use the standard hdf5 storage service, there are no additional arguments
        or keyword arguments to pass!

        '''
        #kwargs['trajectory'] = self

        remove_from_storage = kwargs.pop('remove_from_storage',False)
        remove_empty_groups = kwargs.get('remove_empty_groups',False)

        fetched_items = self._nn_interface._fetch_items(REMOVE, iterable, args, kwargs)

        if fetched_items:
            if self._stored and remove_from_storage:
                try:
                    self._storage_service.store(globally.LIST, fetched_items,
                                               trajectory_name=self.v_name)
                except:
                    self._logger.error('Could not remove >>%s<< from the trajectory. Maybe the'
                                       ' item(s) was/were never stored to disk.')
                    raise

            for msg, item, dummy1, dummy2 in fetched_items:
                self._nn_interface._remove_node_or_leaf(item, remove_empty_groups)

        else:
            self._logger.warning('Your removal was not successful, could not find a single '
                                 'item to remove.')


    def _remove_incomplete_runs(self):
        self._storage_service.store(globally.REMOVE_INCOMPLETE_RUNS, self,
                                      trajectory_name=self.v_name)



    def f_shrink(self):
        ''' Shrinks the trajectory and removes all exploration arrays from the parameters.
        Only possible if the trajectory has not been stored to disk before or was loaded as new.

        :raises: TypeError if the trajectory was stored before.

        '''
        if self._stored:
            raise TypeError('Your trajectory is already stored to disk of database, shrinking is '
                            'not allowed.')

        for key, param in self._explored_parameters:
            param._shrink()

        self._explored_parameters={}

        self._run_information = {}
        self._single_run_ids = {}
        self._add_run_info(0)

    def _preset(self,name,args,kwargs):
        if name in self:
            raise ValueError('Parameter >>%s<< is already part of your trajectory, use the normal'
                             'accessing routine to change config.' % name)
        else:
            self._changed_default_parameters[name] = (args, kwargs)

    def f_preset_config(self, config_name, *args, **kwargs):
        ''' Similar to func:`~pypet.trajectory.Trajectory.f_preset_parameter`.'''
        config_name = 'config' + '.' + config_name
        self._preset(config_name,args, kwargs)


    def f_preset_parameter(self, param_name, *args, **kwargs):
        ''' Can be called before parameters are added to the Trajectory in order to change the
        values that are stored
        into the parameter on creation.

        After creation of a Parameter, the instance of the parameter is called with `param.f_set(*args,**kwargs)`.


        Before an experiment is carried out it is checked if all parameters that were marked were also
        preset.

        :param param_name: The name of the parameter that is to be changed after its creation.

        Example:

        >>> traj.f_preset_parameter('groupA.param1', data=44)
        >>> traj.f_add_parameter('groupA.param1', data=11)
        >>> traj.parameters.groupA.param1.get()
        44

        '''
        param_name = 'parameters' + '.' + param_name
        self._preset(param_name,args, kwargs)

    def _prepare_experiment(self):

        if len(self._changed_default_parameters):
            raise pex.DefaultReplacementError(
                'The following parameters were supposed to replace a '
                'default value, but it was never tried to '
                'add default values with these names: %s' %
                str(self._changed_default_parameters))

        self.f_lock_parameters()
        self.f_lock_derived_parameters()
        for param in self._explored_parameters.itervalues():
            param.set_fullcopy = self._full_copy

        self.f_store()


    def f_find_idx(self, name_list, predicate):
        ''' Finds a given single run index given a particular condition on parameters.

        :param name_list: A list of parameter names the predicate applies to
        :param predicate: A lambda predicate for filtering that evaluates to either true or false.

        :return: a generator returning the matching single run idx

        Example:

        >>> predicate = lambda param1,param2: param1.get()==4 and param2.get() in [1,2]
        >>> iterator = traj.f_find_idx(['groupA.param1','groupA.param2'], predicate)
        >>> [x for x in iterator]
        [0,2,17,36]

        '''
        if not isinstance(name_list, (list,tuple)):
            name_list = [name_list]

        iter_list = []
        for name in name_list:
            param = self.f_get(name)
            if param.v_parameter:
                raise TypeError('>>%s<< is not a parameter it is a %s, find is not applicable' %
                                (name, str(type(param))))

            if param.f_is_array():
                iter_list.append(iter(param.f_get_array()))
            else:
                iter_list.append((param.f_get() for dummy in xrange(len(self))))

        logic_iter = it.imap(predicate, *iter_list)

        for idx, item in enumerate(logic_iter):
            if item:
                yield idx


    def __len__(self):
        if hasattr(self,'_run_information') and len(self._run_information)>0:
            return len(self._run_information)
        else:
            return self._length_during_run


    def __getstate__(self):
        result = self.__dict__.copy()

        if not self.v_full_copy:
            result['_run_information'] ={}
            result['_single_run_ids'] = {}
            result['_length_during_run'] = len(self)

        del result['_logger']
        return result

    def __setstate__(self, statedict):
        self.__dict__.update(statedict)

        #self._tree= NNGroupNode(parent_trajectory=self, predecessors=[], depth=0, name='root')
        self._logger = logging.getLogger('pypet.trajectory.Trajectory=' + self.v_name)



    def _load_class(self, full_class_string):
        """Dynamically f_load a class from a string.
        """

        class_data = full_class_string.split(".")
        module_path = ".".join(class_data[:-1])
        class_str = class_data[-1]

        module = imp.import_module(module_path)
        # Finally, we retrieve the Class
        return getattr(module, class_str)







    def f_is_completed(self, name_or_id=None):
        '''Whether or not a given run is completed.
        
        If no run is specified it is checked whether all runs were completed.
        
        :param name_or_id: Nam or id of a run to check
        
        :return: True or False
        
        '''

        if name_or_id is None:
            return all((runinfo['completed'] for runinfo in self._run_information.itervalues()) )
        else:
            return self.f_get_run_information(name_or_id)['completed']



    def f_expand(self, build_function, *args, **kwargs):
        '''Similar to :func:`~pypet.trajectory.Trajectory.f_explore`, but can be used to enlarge
        already completed trajectories.

        '''
        build_dict = build_function(*args, **kwargs)

        if isinstance(build_dict, tuple):
            build_dict, dummy = build_dict

        if not set(self._explored_parameters.keys()) == set(build_dict.keys()):
            raise TypeError('You have to enlarge dimensions you have explored before! Currently'
                            ' explored parameters are not the ones you specified in your building'
                            ' dictionary, i.e. %s != %s' %
                            (str(set(self._explored_parameters.keys())) ,
                                                           str(set(build_dict.keys()))))

        count = 0#Don't like it but ok
        for key, builditerable in build_dict.items():
            act_param = self.f_get(key, check_uniqueness=True)
            if not act_param.v_leaf or not act_param.v_parameter:
                raise ValueError('%s is not an appropriate search string for a parameter.' % key)


            act_param._expand(builditerable)

            name = act_param.v_full_name

            self._explored_parameters[name] = act_param

            if count == 0:
                length = len(act_param)#Not so nice, but this should always be the same numbert
            else:
                if not length == len(act_param):
                    raise ValueError('The parameters to _explore have not the same size!')

            for irun in range(length):
                self._add_run_info(irun)


    def f_explore(self, build_dict):
        ''' Prepares the trajectory to explore the parameter space.
        To explore the parameter space you need to provide a dictionary with the names of the
        parameters to explore as keys and iterables specifying the exploration array as values.

        All iterables need to have the same length.

        Example:
        >>> traj.explore({'groupA.param1' : [1,2,3,4,5], 'groupA.param2':['a','b','c','d','e']})

        '''

        count = 0#Don't like it but ok
        for key, builditerable in build_dict.items():
            act_param = self.f_get(key, check_uniqueness=True)
            if not act_param.v_leaf or not act_param.v_parameter:
                raise ValueError('%s is not an appropriate search string for a parameter.' % key)

            act_param._explore(builditerable)

            name = act_param.v_full_name

            self._explored_parameters[name] = act_param

            if count == 0:
                length = len(act_param)#Not so nice, but this should always be the same numbert
            else:
                if not length == len(act_param):
                    raise ValueError('The parameters to _explore have not the same size!')

            for irun in range(length):
                self._add_run_info(irun)


    def _add_run_info(self, idx):

        runname = FORMATTED_RUN_NAME % idx
        self._single_run_ids[runname] = idx
        self._single_run_ids[idx] = runname
        info_dict = {}

        info_dict['idx'] = idx
        info_dict['timestamp'] = 42.0
        info_dict['time'] = '~waste ur t1me with a phd'
        info_dict['completed'] = 0
        info_dict['name'] = runname
        info_dict['parameter_summary'] = 'Not yet my friend!'

        self._run_information[runname] = info_dict


    def f_lock_parameters(self):
        '''Locks all parameters'''
        for key, par in self._parameters.items():
            par.f_lock()

    def f_lock_derived_parameters(self):
        '''Locks all derived parameters'''
        for key, par in self._derived_parameters.items():
            par.f_lock()

    def _finalize(self):
        self.f_restore_default()
        self._nn_interface._change_root(self)

    def f_update_skeleton(self):
        '''Loads the full skeleton from the storage service.

        This needs to be done after a successful exploration in order to update the trajectory tree
        with all results and derived parameters from the individual single runs.
        This will only add empty results and derived parameters (i.e. the skeleton) and load annotations.

        '''
        self.f_load(self.v_name,None, False, globally.UPDATE_SKELETON, globally.UPDATE_SKELETON,
                  globally.UPDATE_SKELETON)

    def f_load_item(self,item,*args,**kwargs):
        '''Loads a single item, see also :func:`~pypet.trajectory.Trajectory.f_load_items`'''
        self.f_load_items([item],*args,**kwargs)

    def f_load_items(self, iterator, *args, **kwargs):
        ''' Loads parameters specified in `iterator`. You can directly list the Parameter objects or their
        names.

        If names are given the `~pypet.trajectory.Trajectory.f_get` method is applied to find the
        parameters or results in the
        trajectory.
        If kwargs f_contains the keyword >>only_empties=True<<, only empty parameters or
        results are passed to the
        storage service to get loaded.

        This function is useful if you called :func:`~pypet.trajectory.Trajectory.f_update_skeleton`
        before and now you want
        to load the data of individual results one by one.

        :param iterator: A list with parameters or results to be loaded.

        :param args: Additional arguments directly passed to the storage service

        :param kwargs:

            Additional keyword arguments directly passed to the storage service (except the kwarg
            only_empties)

            If you use the standard hdf5 storage service, you can pass the following additional
            keyword argument:

            :param load_only:

                If you load a result, you can partially load it and ignore the rest.
                Just specify the
                name of the data you want to load. You can also provide a list,
                for example `load_only='spikes'`, `load_only=['spikes','membrane_potential']`

                Throws a ValueError if data cannot be found.

        '''

        if not self._stored:
            raise TypeError(
                'Cannot f_load stuff from disk for a trajectory that has never been stored.')

        #only_empties = kwargs.pop('only_empties', False)

        fetched_items = self._nn_interface._fetch_items(LOAD, iterator, args, kwargs)
        if fetched_items:
            self._storage_service.load(globally.LIST, fetched_items,
                                      trajectory_name=self.v_name)
        else:
            self._logger.warning('Your loading was not successful, could not find a single item '
                                 'to f_load.')


    def f_load(self,
             trajectory_name=None,
             trajectory_index = None,
             as_new=False,
             load_parameters=None,
             load_derived_parameters=None,
             load_results=None):
        ''' Loads a trajectory via the storage service.

        :param trajectory_name:

            Name of the trajectory to be loaded. If no name or index is specified
            the current name of the trajectory is used.

        :param trajectory_index:

            If you don't specify a name you can also specify an index.
            The corresponding trajectory in the hdf5 file at the index
            position is loaded.

        :param as_new:

            Whether you want to rerun the experiments. So the trajectory is loaded only
            with parameters, the current trajectory name is kept in this case, which should be different
            from the trajectory name specified in the input parameter `trajectory_name`.
            If you load `as_new=True` all parameters and derived parameters are unlocked.
            If you load `as_new=False` the current trajectory is replaced by the one on disk,
            i.e. name, timestamp, formatted time etc. are all taken from disk.

        :param load_parameters: How parameters and config items are loaded

        :param load_derived_parameters: How derived parameters are loaded

        :param load_results: How results are loaded

            You can specify how to load the parameters and config/derived_parameters/results.

                :const:`pypet.globally.LOAD_NOTHING`: (0)

                    Nothing is loaded

                :const:`pypet.globally.LOAD_SKELETON`: (1)

                        The skeleton including annotations are loaded, i.e. the items are empty.
                        Note that if the items already exist in your trajectory an Attribute
                        Error is thrown. If this is the case use -1 instead.

                :const:`pypet.globally.LOAD_DATA`: (2)

                    The whole data is loaded.
                    Note that if the items already exist in your trajectory an Attribute
                    Error is thrown. If this is the case use -2 instead.


                :const:`pypet.globally.UPDATE_SKELETON`: (-1)

                    The skeleton and annotations are updated, i.e. only items that are not currently part
                    of your trajectory are loaded empty

                :const:`pypet.globally.UPDATE_DATA`: (-2) Like (2)

                    Only items that are currently not in your trajectory are loaded with data.

        :raises:

            Attribute Error:

                If options 1 and 2 (load skeleton and load data) are applied but the
                objects already exist in your trajectory. This prevents implicitly overriding
                data in RAM.
                Use -1 and -2 instead to load
                only items that are currently not in your trajectory in RAM.
                Or remove the items you want to 'reload' first.

        '''

        if trajectory_name is None and trajectory_index is None:
            trajectory_name = self.v_name

        if as_new and load_parameters is None:
            load_parameters=globally.LOAD_DATA
        elif load_parameters is None:
            load_parameters = globally.UPDATE_DATA

        if as_new and load_derived_parameters is None:
            load_derived_parameters = globally.LOAD_NOTHING
        elif load_derived_parameters is None:
            load_derived_parameters=globally.LOAD_SKELETON

        if as_new and load_results is None:
            load_results = globally.LOAD_NOTHING
        elif load_results is None:
            load_results = globally.LOAD_SKELETON

        self._storage_service.load(globally.TRAJECTORY, self, trajectory_name=trajectory_name,
                                  trajectory_index=trajectory_index,
                                  as_new=as_new, load_params=load_parameters,
                                  load_derived_params=load_derived_parameters,
                                  load_results=load_results)

        if as_new:
            for param in self._parameters.itervalues():
                param.f_unlock()

            for param in self._derived_parameters.itervalues():
                param.f_unlock()


    def _check_if_both_have_same_parameters(self, other_trajectory,
                                            ignore_trajectory_derived_parameters):

        assert isinstance(other_trajectory, Trajectory)
        self.f_update_skeleton()
        other_trajectory.f_update_skeleton()

        ## Load all parameters of the current and the given Trajectory
        if self._stored:
            self.f_load_items(self._parameters.values(), only_empties=True)
        if other_trajectory._stored:
            other_trajectory.f_load_items(other_trajectory._parameters.values(), only_empties=True)

        self.f_restore_default()
        other_trajectory.f_restore_default()

        allmyparams = self._parameters.copy()
        allotherparams = other_trajectory._parameters.copy()

        if not ignore_trajectory_derived_parameters and 'derived_parameters.trajectory' in self:
            my_traj_dpars = self.f_get('derived_parameters.trajectory').f_to_dict()
            allmyparams.update(my_traj_dpars)
            other_traj_dpars = other_trajectory.f_get('derived_parameters.trajectory').f_to_dict()
            allotherparams.update(other_traj_dpars)


        ## Check if the trajectories have the same parameters:
        my_keyset = set(allmyparams.keys())
        other_keyset = set(allotherparams.keys())

        if not my_keyset == other_keyset:
            diff1 = my_keyset - other_keyset
            diff2 = other_keyset - my_keyset
            raise TypeError('Cannot f_merge trajectories, they do not live in the same space,the '
                            'f_set of parameters >>%s<< is only found in the current trajectory '
                            'and >>%s<< only in the other trajectory.' % (str(diff1), str(diff2)))

        for key, other_param in allotherparams.items():
            my_param = self.f_get(key)
            if not my_param._values_of_same_type(my_param.f_get(), other_param.f_get()):
                raise TypeError('Cannot f_merge trajectories, values of parameters >>%s<< are not '
                                'of the same type. Types are %s (current) and %s (other).' %
                                (key, str(type(my_param.f_get())), str(type(other_param.f_get()))))


    def f_backup(self, backup_filename):
        '''Backs up the trajectory with the given storage service.

        :param backup_filename: Name of file where to store the backup!

        '''
        self._storage_service.store(globally.BACKUP, self, trajectory_name=self.v_name,
                                    backup_filename=backup_filename)


    def f_merge(self, other_trajectory, trial_parameter=None, remove_duplicates = False,
                ignore_trajectory_derived_parameters = False,
                ignore_trajectory_results= False,
                backup_filename = None,
                move_nodes=False,
                delete_trajectory=False):
        ''' Merges another trajectory into the current trajectory.

        Both trajectories must live in the same space. That means both need to have the same
        parameters with similar types of values. Yet, if the other trajectory explores
        different parameters than the current one, merging will take care of that.

        :param other_trajectory: Other trajectory instance to merge into the current one.

        :param trial_parameter:

            If you have a particular parameter that specifies only the trial
            number, i.e. an integer parameter running form 0 to N1 and
            0 to N2, the parameter is modified that after merging it will
            cover the range 0 to N1+N2.

        :param remove_duplicates:

            Whether you want to remove duplicate parameter points.
            Requires N1 * N2 (quadratic complexity in single runs).

        :param ignore_trajectory_derived_parameters:

                    Whether you want to ignore or merge derived parameters
                    kept under `.derived_parameters.trajectory`

        :param ignore_trajectory_results:

             As above but with results. If you have trajectory results
             with the same name in both trajectories, the result
             in the current trajectory is kept and the other one is
             not merged into the current trajectory.

        :param backup_filename:

            If specified, backs up both trajectories into the given filename.
            You could also say backup_filename = True, than the trajectories
            are backed up into a file in your data folder and a name is
            automatically chosen.

        :param move_nodes:

           If you use the HDF5 storage service and both trajectories are
           stored in the same file merging is achieved fast directly within
           the file. You can choose if you want to copy nodes from the other
           trajectory to the current one, or if you want to move them. Accordingly
           the stored data is no longer accessible in the other trajectory.

        :param delete_trajectory: If you want to delete the other trajectory after merging

        If you cannot directly merge trajectories within one HDF5 file a slow merging process
        is used. Results are loaded, stored and emptied again one after the other. Might take
        some time!

        Annotations of parameters and derived parameters under `.derived_parameters.trajectory`
        are NOT merged. If you wish to extract the annotations of these parameters you have to
        do that manually before merging. Note that annotations of results and derived parameters
        of single runs are copied, so you don't have to worry about these.

        '''

        ## Check if trajectories can be merged
        self._check_if_both_have_same_parameters(other_trajectory,
                                                 ignore_trajectory_derived_parameters)

        ## BACKUP if merge is possible
        if not backup_filename is None:
            if backup_filename==1 or backup_filename==True:
                backup_filename = None

            other_trajectory.f_backup(backup_filename=backup_filename)
            self.f_backup(backup_filename=backup_filename)



        self._logger.info('Merging the parameters.')
        used_runs, changed_parameters = self._merge_parameters(other_trajectory, remove_duplicates,
                                                               trial_parameter,
                                                               ignore_trajectory_derived_parameters)

        if np.all(used_runs == 0):
            self._logger.warning('Your f_merge discards all runs of the other trajectory, maybe you '
                                 'try to f_merge a trajectory with itself?')
            return
            # if not ignore_trajectory_derived_parameters and 'derived_parameters.trajectory' in other_trajectory:
        #     self._logger.info('Merging derived trajectory parameters')
        #     changed_derived_parameters = self._merge_trajectory_derived_parameters(other_trajectory,used_runs)
        # else:
        #     changed_derived_parameters = []

        rename_dict = {}

        if not ignore_trajectory_results and 'results.trajectory' in other_trajectory:
            self._logger.info('Merging trajectory results skeletons.')
            self._merge_trajectory_results(other_trajectory, rename_dict)

        self._logger.info('Merging single run skeletons.')
        self._merge_single_runs(other_trajectory, used_runs, rename_dict)

        #adding_length = sum(used_runs)



        self._logger.info('Start copying results and single run derived parameters.')

        self._logger.info('Updating Trajectory information and changed parameters in storage.')

        self._storage_service.store(globally.PREPARE_MERGE, self,
                                   trajectory_name=self.v_name,
                                   changed_parameters=changed_parameters,
                                   rename_dict=rename_dict)

        try:
            self._storage_service.store(globally.MERGE, None, trajectory_name=self.v_name,
                                       other_trajectory_name=other_trajectory.v_name,
                                       rename_dict=rename_dict, move_nodes=move_nodes,
                                       delete_trajectory=delete_trajectory)



        except pex.NoSuchServiceError:
            self._logger.warning('My storage service does not support merging of trajectories, '
                                 'I will use the f_load mechanism of the other trajectory and copy '
                                 'the results manually and slowly. Note that thereby the other '
                                 'trajectory will be altered.')


            self._merge_slowly(other_trajectory, rename_dict)

        except ValueError, e:
            self._logger.warning(str(e))

            self._logger.warning('I will use the f_load mechanism of the other trajectory and copy '
                                 'the results manually and slowly. Note that thereby the other '
                                 'trajectory will be altered.')

            self._merge_slowly(other_trajectory, rename_dict)

        self._logger.info('Finished Merging!')


    def _merge_slowly(self, other_trajectory, rename_dict):



        for other_key, new_key in rename_dict.iteritems():

            other_instance = other_trajectory.f_get(other_key)
            was_empty = False
            if other_instance.f_is_empty():
                was_empty = True
                other_trajectory.f_load_items(other_instance)

            my_instance = self.f_get(new_key)
            if not my_instance.f_is_empty():
                raise RuntimeError('You want to slowly f_merge results, but your target result '
                                   '>>%s<< is not _empty, this should not happen.' %
                                   my_instance.v_full_name)

            load_dict = other_instance._store()
            my_instance._load(load_dict)
            my_instance.f_set_annotations(**other_instance.v_annotations.f_to_dict(copy=False))

            self.f_store_item(my_instance)

            if was_empty:
                other_instance._empty()
                my_instance._empty()


    def _merge_trajectory_results(self, other_trajectory, rename_dict):


        other_results = other_trajectory.f_get('results.trajectory').f_to_dict()

        for key, result in other_results.iteritems():


            if key in self._results:
                self._logger.warning('You already have a trajectory result called >>%s<< in your '
                                     'trajectory. I will not copy it.' % key)
                continue

            rename_dict[key] = key
            comment = result.v_comment
            result_type = result.f_get_class_name()
            result_type = self._create_class(result_type)
            self.f_add_result(result_type,key, comment=comment)

    def _merge_single_runs(self, other_trajectory, used_runs, rename_dict):

        count = len(self)
        runnames = other_trajectory.f_get_run_names()
        for runname in runnames:
            idx = other_trajectory.f_get_run_information(runname)['idx']
            if used_runs[idx]:
                try:
                    results = other_trajectory.f_get('results.' + runname).f_to_dict()
                except AttributeError:
                    results = {}

                try:
                    derived_params = other_trajectory.f_get(
                        'derived_parameters.' + runname).f_to_dict()
                except AttributeError:
                    derived_params = {}

                time = other_trajectory.f_get_run_information(runname)['time']
                timestamp = other_trajectory.f_get_run_information(runname)['timestamp']
                completed = other_trajectory.f_get_run_information(runname)['completed']

                new_runname = FORMATTED_RUN_NAME % count

                self._run_information[new_runname] = dict(idx=count,
                                                          time=time, timestamp=timestamp,
                                                          completed=completed)

                self._single_run_ids[count] = new_runname
                self._single_run_ids[new_runname] = count

                count += 1

                for result_name, result in results.iteritems():
                    new_result_name = self._rename_key(result_name, 1, new_runname)
                    rename_dict[result_name] = new_result_name
                    comment = result.v_comment
                    result_type = result.f_get_class_name()
                    result_type = self._create_class(result_type)
                    self.f_add_result(result_type,new_result_name, comment=comment)

                for dpar_name, dpar in derived_params.iteritems():
                    new_dpar_name = self._rename_key(dpar_name, 1, new_runname)
                    rename_dict[dpar_name] = new_dpar_name
                    comment = dpar.v_comment
                    param_type = dpar.f_get_class_name()
                    param_type = self._create_class(param_type)
                    self.f_add_derived_parameter(param_type,new_dpar_name, comment=comment)


            else:
                continue


    def _rename_key(self, key, pos, newname):
        split_key = key.split('.')
        split_key[pos] = newname
        renamed_key = '.'.join(split_key)
        return renamed_key

    def _merge_parameters(self, other_trajectory, remove_duplicates=False, trial_parameter=None,
                          ignore_trajectory_derived_parameters=False):


        if trial_parameter:
            if remove_duplicates:
                self._logger.warning('You have given a trial parameter and you want to '
                                     'f_remove_items duplicates. There cannot be any duplicates '
                                     'when adding trials, I will not look for duplicates.')
                remove_duplicates = False

        if trial_parameter:
            my_trial_parameter = self.f_get(trial_parameter)
            other_trial_parameter = other_trajectory.f_get(trial_parameter)
            if not isinstance(my_trial_parameter, BaseParameter):
                raise TypeError('Your trial_parameter >>%s<< does not evaluate to a real parameter'
                                ' in the trajectory' % trial_parameter)

            if my_trial_parameter.f_is_array():
                my_trial_list = my_trial_parameter.f_get_array()
            else:
                my_trial_list = [my_trial_parameter.f_get()]

            if other_trial_parameter.f_is_array():
                other_trial_list = other_trial_parameter.f_get_array()
            else:
                other_trial_list = [other_trial_parameter.f_get()]
            mytrialset = set(my_trial_list)
            mymaxtrial = max(mytrialset)

            if mytrialset != set(range(mymaxtrial + 1)):
                raise TypeError('In order to specify a trial parameter, this parameter must '
                                'contain integers from 0 to %d, but it infact it '
                                'f_contains >>%s<<.' % (mymaxtrial, str(mytrialset)))

            othertrialset = set(other_trial_list)
            othermaxtrial = max(othertrialset)
            if othertrialset != set(range(othermaxtrial + 1)):
                raise TypeError('In order to specify a trial parameter, this parameter must '
                                'contain integers from 0 to %d, but it infact it f_contains >>%s<< in the other trajectory.' % (
                    othermaxtrial, str(othertrialset)))

            trial_parameter = my_trial_parameter.v_full_name

            if not trial_parameter in self._explored_parameters:
                self._explored_parameters[trial_parameter] = my_trial_parameter


        ## Check which parameters differ:
        params_to_change = {}

        params_to_merge = other_trajectory._parameters.copy()

        if not ignore_trajectory_derived_parameters and 'derived_parameters.trajectory' in self:
            trajectory_derived_parameters = other_trajectory.f_get(
                'derived_parameters.trajectory').f_to_dict()
            params_to_merge.update(trajectory_derived_parameters)

        for key, other_param in params_to_merge.iteritems():

            my_param = self.f_get(key)
            if not my_param._values_of_same_type(my_param.f_get(), other_param.f_get()):
                raise TypeError('The parameters with name >>%s<< are not of the same type, cannot '
                                'f_merge trajectory.' % key)

            if my_param.v_full_name == trial_parameter:
                params_to_change[key] = (my_param, other_param)
                continue

            if (my_param.f_is_array()
                or other_param.f_is_array()
                or not my_param._equal_values(my_param.f_get(), other_param.f_get())):

                params_to_change[key] = (my_param, other_param)
                if not my_param.f_is_array() and not other_param.f_is_array():
                    remove_duplicates = False

        ## Now first check if we use all runs ore f_remove_items duplicates:
        use_runs = np.ones(len(other_trajectory))
        if remove_duplicates:

            for irun in xrange(len(other_trajectory)):
                for jrun in xrange(len(self)):
                    change = True
                    for my_param, other_param in params_to_change.itervalues():
                        if other_param.f_is_array():
                            other_param.f_set_parameter_access(irun)

                        if my_param.f_is_array():
                            my_param.f_set_parameter_access(jrun)

                        val1 = my_param.f_get()
                        val2 = other_param.f_get()

                        if not my_param._equal_values(val1, val2):
                            change = False
                            break
                    if change:
                        use_runs[irun] = 0.0
                        break

            ## Restore changed default values
            for my_param, other_param in params_to_change.itervalues():
                other_param.f_restore_default()
                my_param.f_restore_default()


        ## Now f_merge into the new trajectory
        adding_length = int(sum(use_runs))
        if adding_length == 0:
            return 0, []

        for my_param, other_param in params_to_change.itervalues():
            fullname = my_param.v_full_name

            if fullname == trial_parameter:
                other_array = [x + mymaxtrial + 1 for x in other_trial_list]
            else:
                if other_param.f_is_array():
                    other_array = (x for run, x in it.izip(use_runs, other_param.f_get_array()) if \
                                   run)
                else:
                    other_array = (other_param.f_get() for dummy in xrange(adding_length))

            if not my_param.f_is_array():
                my_param.f_unlock()
                my_param._explore((my_param.f_get() for dummy in xrange(len(self))))

                #self._explored_parameters[my_param.v_full_name]=my_param

            my_param.f_unlock()

            my_param._expand(other_array)

            if not fullname in self._explored_parameters:
                self._explored_parameters[fullname] = my_param

        return use_runs, params_to_change.keys()


    def f_store(self):
        ''' Stores the trajectory to disk.

        If you use the HDF5 Storage Service only novel data is stored to disk.

        If you have results that have been stored to disk before only new items are added and
        already present data is NOT overwritten.

        Overwriting existing data with the HDF5 storage service is currently not supported.

        '''
        #self._srn_add_explored_params()
        self._storage_service.store(globally.TRAJECTORY, self, trajectory_name=self.v_name)
        self._stored = True

    def f_is_empty(self):
        '''Whether no results nor parameters have been added yet to the trajectory (ignores config).'''
        return (len(self._parameters) == 0 and
                len(self._derived_parameters) == 0 and
                len(self._results) == 0)




    def f_restore_default(self):
        '''Restores the default value in all explored parameters'''
        for param in self._explored_parameters.itervalues():
            param.f_restore_default()


    def f_prepare_parameter_space_point(self, n):
        ''' Notifies the explored parameters what the current point in the parameter space it is,
        i.e. which is the current run.

        The function calls
        :func:`~pypet.parameter.Parameter.set_parameter_access` in all its explored parameters.

        '''

        # extract only one particular paramspacepoint
        for key, val in self._explored_parameters.items():
            val.f_set_parameter_access(n)


    def _make_single_run(self, idx):
        ''' Creates a SingleRun object for parameter exploration.
        '''

        for param in self._explored_parameters.itervalues():
            param.f_set_parameter_access(idx)

        name = self.f_idx_to_run(idx)
        return SingleRun(name, idx, self)







__author__ = 'Robert Meyer'

''' This module contains constants defined on a global scale,
that might be interesting to the user.
'''


import numpy as np




###################### Supported Data ########################


PARAMETERTYPEDICT={"<type 'bool'>": bool,
 "<type 'complex'>": complex,
 "<type 'float'>": float,
 "<type 'int'>": int,
 "<type 'numpy.bool_'>": np.bool_,
 "<type 'numpy.complex128'>": np.complex128,
 "<type 'numpy.complex64'>": np.complex64,
 "<type 'numpy.float32'>": np.float32,
 "<type 'numpy.float64'>": np.float64,
 "<type 'numpy.int16'>": np.int16,
 "<type 'numpy.int32'>": np.int32,
 "<type 'numpy.int64'>": np.int64,
 "<type 'numpy.int8'>": np.int8,
 "<type 'numpy.string_'>": np.string_,
 "<type 'numpy.uint16'>": np.uint16,
 "<type 'numpy.uint32'>": np.uint32,
 "<type 'numpy.uint64'>": np.uint64,
 "<type 'numpy.uint8'>": np.uint8,
 "<type 'str'>": str}
''' A Mapping (dict) from the the string represenation of a type and the type.

These are the so far supported types of the storage service and the standard parameter!
'''


PARAMETER_SUPPORTED_DATA = (np.int8,
                       np.int16,
                       np.int32,
                       np.int64,
                       np.int,
                       np.int_,
                       np.uint8,
                       np.uint16,
                       np.uint32,
                       np.uint64,
                       np.bool,
                       np.bool_,
                       np.float32,
                       np.float64,
                       np.float,
                       np.float_,
                       np.complex64,
                       np.complex,
                       np.complex_,
                       np.str,
                       np.str_)
'''Set of supported scalar types by the storage and the standard parameter'''


################### HDF5 Naming and Comments ##########################


HDF5_STRCOL_MAX_NAME_LENGTH = 64
'''Maximum length of a (short) name'''
HDF5_STRCOL_MAX_LOCATION_LENGTH = 128
'''Maximum lenght of the location'''
HDF5_STRCOL_MAX_VALUE_LENGTH = 64
''' Maximum length of a value string'''
HDF5_STRCOL_MAX_COMMENT_LENGTH = 256
''' Maximum length of a comment '''
HDF5_STRCOL_MAX_ARRAY_LENGTH = 2048
''' Maximum length of a parameter array summary '''

######## Multiprocessing Modes #############

WRAP_MODE_QUEUE = 'QUEUE'
'''For multiprocesseing, queue multiprocessing mode '''
WRAP_MODE_LOCK = 'LOCK'
''' Lock multiprocessing mode '''
WRAP_MODE_NONE = 'NONE'
''' No multiprocessing wrapping for the storage service'''


############ Loading Constants ###########################
LOAD_SKELETON = 1
'''For trajectory loading, loads only the skeleton '''
LOAD_DATA = 2
''' Loads skeleton and data '''
LOAD_NOTHING = 0
''' Loads nothing '''
UPDATE_SKELETON = -1
''' Updates skeleton, i.e. adds only items that are not part of your current trajectory '''
UPDATE_DATA = -2
''' Updates skeleton and data, adds only items that are not part of your current trajectory'''
# LOAD_ANNOTATIONS = 3
# ''' Reloads all annotations! Be Aware this option overrides all given annotations!'''




##################### STORING Message Constants ################################
LEAF ='LEAF'
''' For trajectory or item storage, stores a *leaf* node, i.e. parameter or result object'''
UPDATE_LEAF = 'UPDATE_LEAF'
''' Updates a *leaf* node, currently only parameters that are extended in length can be updated'''
TRAJECTORY = 'TRAJECTORY'
''' Stores the whole trajectory'''
MERGE = 'MERGE'
''' Merges two trajectories '''
GROUP = 'GROUP'
''' Stores a group node, can be recursive'''
LIST = 'LIST'
''' Stores a list of different things, in order to avoid reopening and closing of the hdf5 file'''
SINGLE_RUN = 'SINGLE_RUN'
''' Stores a single run'''
PREPARE_MERGE = 'PREPARE_MERGE'
''' Updates a trajectory before it is going to be merged'''
BACKUP = 'BACKUP'
''' Backs up a trajectory'''
REMOVE='REMOVE'
''' Removes an item from hdf5 file'''
REMOVE_INCOMPLETE_RUNS = 'REMOVE_INCOMPLETE_RUNS'
''' Removes incomplete runs to continue a crashed trajectory'''
TREE = 'TREE'
''' Stores a subtree of the trajectory'''


# Search Strategies
BFS = 'BFS'
'''For search in trajectory tree, breadth first search, default strategy'''
DFS = 'DFS'
'''Depth first search in trajectory tree, not recommended'''



__author__ = 'Robert Meyer'


import numpy as np
import unittest
from pypet.brian.parameter import BrianParameter
from pypet.tests.parameter_test import ParameterTest, Dummy
from brian.stdunits import mV, mA, kHz,ms


class BrianParameterTest(ParameterTest):

    def setUp(self):

        if not hasattr(self,'data'):
            self.data = Dummy()
        self.data.mV1 = 1*mV
        self.data.ampere1 = 1*mA
        self.data.msecond17 = 16*ms
        self.data.kHz05 = 0.5*kHz


        super(BrianParameterTest,self).setUp()


    def make_params(self):
        self.param = Dummy()
        for key, val in self.data.__dict__.items():
            self.param.__dict__[key] = BrianParameter(self.location+'.'+key, val, comment=key)



    def explore(self):
        self.explore_dict={'npstr':[np.array(['Uno', 'Dos', 'Tres']),
                               np.array(['Cinco', 'Seis', 'Siette']),
                            np.array(['Ocho', 'Nueve', 'Diez'])],
                           'val0':[1,2,3],
                           'mV1' : [42.0*mV,3*mV,4*mV]}




        ## Explore the parameter:
        for key, vallist in self.explore_dict.items():
            self.param.__dict__[key]._explore(vallist)


if __name__ == '__main__':
    unittest.main()
     #cProfile.run('unittest.main()',sort=0)




'''
Created on 24.05.2013
'''
import itertools as itools



def cartesian_product(parameter_dict, combined_parameters = ()):
    ''' Generates a Cartesian product of the input parameter dictionary.

    For example:
    >>> print cartesian_product({'param1':[1,2,3], 'param2':[42.0, 52.5]})
    {'param1':[1,1,2,2,3,3],'param2': [42.0,52.5,42.0,52.5,42.0,52.5]}

    :param param_dict:

        Dictionary containing parameter names as keys and iterables of data to explore.

    :param combined_parameter_list:

        Tuple of tuples. Defines the order of the parameters and parameters that are
        linked together.
        If an inner tuple contains only a single item, you can space the
        inner tuple brackets.


        For example:
        >>> print cartesian_product( {'param1': [42.0, 52.5], 'param2':['a', 'b'],\
        'param3' : [1,2,3]}, (('param3',),('param1', 'param2')))
        {param3':[1,1,2,2,3,3],'param1' : [42.0,52.5,42.0,52.5,42.0,52.5],\
        'param2':['a','b','a','b','a','b']}

    :returns: Dictionary with cartesian product lists.

    '''
    if not combined_parameters:
        combined_parameters = list(parameter_dict)
    else:
        combined_parameters = list(combined_parameters)

    for idx,item in enumerate(combined_parameters):
        if isinstance(item,basestring):
            combined_parameters[idx]= (item,)

    iterator_list = []
    for item_tuple in combined_parameters:
        inner_iterator_list = [parameter_dict[key] for key in item_tuple]
        zipped_iterator = itools.izip(*inner_iterator_list)
        iterator_list.append(zipped_iterator)

    result_dict ={}
    for key in parameter_dict:
        result_dict[key]=[]

    cartesian_iterator = itools.product(*iterator_list)

    for cartesian_tuple in cartesian_iterator:
        for idx, item_tuple in enumerate(combined_parameters):
            for inneridx,key in enumerate(item_tuple):
                result_dict[key].append(cartesian_tuple[idx][inneridx])

    return result_dict




def cartesian_product_old(param_dict, combined_parameter_list=[]):
    ''' Generates a Cartesian product of the input parameters.

    
    For example:
    >>> print cartesian_product({'param1.entry1':[1,2,3], 'param2.entry1' = [42.0, 42.5]})
    {'param1.entry1':[1,1,2,2,3,3],'param2.entry2' = [42.0,42.5,42.0,42.5,42.0,42.5]}
    
    :param param_dict:

        Dictionary containing parameter names as keys and lists of data to explore

    :param combined_parameter_list:

        List of tuples. Defines the order of the parameters and parameters that are linked together.

    :returns: Dictionary with cartesian product lists.

    '''
    if not combined_parameter_list:
        combined_parameter_list = list(param_dict)
    
    for key in param_dict:
        inlist = False
        for ptuple in combined_parameter_list:
            if isinstance(ptuple,basestring):
                ptuple=(ptuple,)

            if key in ptuple:
                inlist = True
                break
        
        if not inlist:
            combined_parameter_list.append((key,))
                    
    
    idx_lists = []
    for idx, paramtuple in enumerate(combined_parameter_list):
        if not isinstance(paramtuple, tuple):
            combined_parameter_list[idx] = (combined_parameter_list[idx],)
            paramtuple = combined_parameter_list[idx]
            
        for idx, param in enumerate(paramtuple):
            value_list = param_dict[param]
            # if not isinstance(value_list, list):
            #     raise TypeError(param + ' is not a list, why should I _explore it anyway?')
            
            if idx == 0:
                param_length = len(value_list)
                idx_lists.append(range(param_length))
            else:
                if not param_length == len(value_list):
                    raise TypeError('Lists of two combined parameters are unequal!')

    
    product_lists = list(itools.product(*idx_lists))
    zipped_lists = zip(*product_lists)
    
    result_dict={}
    result_combined_list=[[]]
    for idx, ituple in enumerate(zipped_lists):
        for param in combined_parameter_list[idx]:
            result_dict[param] = []
            result_combined_list[0].append(param)
            for tupidx in ituple:
                result_dict[param].append(param_dict[param][tupidx])
    
    # result_combined_list[0]=tuple(result_combined_list[0])
    return result_dict
            
                
        



__author__ = 'robert'


import unittest
from pypet.utils.explore import cartesian_product
from pypet.utils.helpful_functions import nested_equal

class CartesianTest(unittest.TestCase):

    def test_cartesian_product(self):

        cartesian_dict=cartesian_product({'param1':[1,2,3], 'param2':[42.0, 52.5]},
                                          ('param1','param2'))
        result_dict = {'param1':[1,1,2,2,3,3],'param2': [42.0,52.5,42.0,52.5,42.0,52.5]}

        self.assertTrue(nested_equal(cartesian_dict,result_dict), '%s != %s' %
                                                        (str(cartesian_dict),str(result_dict)))

        cartesian_dict=cartesian_product( {'param1': [42.0, 52.5], 'param2':['a', 'b'],\
            'param3' : [1,2,3]}, (('param3',),('param1', 'param2')))

        result_dict={'param3':[1,1,2,2,3,3],'param1' : [42.0,52.5,42.0,52.5,42.0,52.5],
                      'param2':['a','b','a','b','a','b']}

        self.assertTrue(nested_equal(cartesian_dict,result_dict), '%s != %s' %
                                                    (str(cartesian_dict),str(result_dict)))




if __name__ == '__main__':
    unittest.main()


from traits.trait_types import self

__author__ = 'Robert Meyer'

import numpy as np
import unittest
from pypet.parameter import Parameter, PickleParameter, BaseResult, ArrayParameter, PickleResult, BaseParameter
from pypet.trajectory import Trajectory, SingleRun
from pypet.storageservice import LazyStorageService
from pypet.utils.explore import cartesian_product
from pypet.environment import Environment
from pypet.storageservice import HDF5StorageService
from pypet import globally
import pickle
import logging
import cProfile
from pypet.utils.helpful_functions import flatten_dictionary
import scipy.sparse as spsp
import os
import shutil
import pandas as pd
from pypet.utils.comparisons import results_equal,parameters_equal
from pypet.utils.helpful_functions import nested_equal

## Removes all the files again to clean up after the tests
REMOVE = True


def simple_calculations(traj, arg1, simple_kwarg):


        all_mat = traj.csc_mat + traj.lil_mat + traj.csr_mat
        Normal_int= traj.Normal.int
        Sum= np.sum(traj.Numpy.double)

        result_mat = all_mat * Normal_int * Sum * arg1 * simple_kwarg



        my_dict = {}

        my_dict2={}
        my_dict3={}
        my_dict4={}
        for key, val in traj.f_to_dict(fast_access=True,short_names=False).items():
            newkey = key.replace('.','_')
            my_dict[newkey] = str(val)
            my_dict2[newkey] =(str(val)+' juhu!')
            my_dict3[newkey] =[str(val)+' juhuhu!']
            my_dict4[newkey]= np.array([str(val)+' Woopieee!'])

        keys = traj.f_to_dict(short_names=False).keys()
        for idx,key in enumerate(keys):
            keys[idx] = key.replace('.','_')

        traj.f_add_result('List.Of.Keys', dict1=my_dict, dict2=my_dict2, dict3=my_dict3,dict4=my_dict4)
        traj.f_add_result('DictsNFrame', keys=keys)
        traj.f_add_result('ResMatrix',result_mat.todense())
        traj.f_add_derived_parameter('All.To.String', str(traj.f_to_dict(fast_access=True,short_names=False)))

        myframe = pd.DataFrame(data ={'TC1':[1,2,3],'TC2':['Waaa',np.nan,''],'TC3':[1.2,42.2,np.nan]})

        traj.DictsNFrame.f_set(myframe)

        traj.f_add_result('IStore.SimpleThings',1.0,3,np.float32(5.0), 'Iamstring',(1,2,3),[4,5,6],zwei=2)

        traj.f_add_result(PickleResult, 'PickleTerror',  test=traj.SimpleThings).f_set_annotations(
            *[1,'hhhhh'],
            **{'peter':np.array([1,2,3,4]),'cat':'fish',
                                     'kkk' : np.array(['1','2','sdfsf'])})

class EnvironmentTest(unittest.TestCase):


    def set_mode(self):
        self.mode = 'LOCK'
        self.multiproc = False
        self.ncores = 1

    def _create_param_dict(self):
        self.param_dict = {}

        self.param_dict['Normal'] = {}
        self.param_dict['Numpy'] = {}
        self.param_dict['Sparse'] ={}
        self.param_dict['Numpy_2D'] = {}
        self.param_dict['Numpy_3D'] = {}
        self.param_dict['Tuples'] ={}

        normal_dict = self.param_dict['Normal']
        normal_dict['string'] = 'Im a test string!'
        normal_dict['int'] = 42
        normal_dict['double'] = 42.42
        normal_dict['bool'] =True
        normal_dict['trial'] = 0

        numpy_dict=self.param_dict['Numpy']
        numpy_dict['string'] = np.array(['Uno', 'Dos', 'Tres'])
        numpy_dict['int'] = np.array([1,2,3,4])
        numpy_dict['double'] = np.array([1.0,2.0,3.0,4.0])
        numpy_dict['bool'] = np.array([True,False, True])

        self.param_dict['Numpy_2D']['double'] = np.array([[1.0,2.0],[3.0,4.0]])
        self.param_dict['Numpy_3D']['double'] = np.array([[[1.0,2.0],[3.0,4.0]],[[3.0,-3.0],[42.0,41.0]]])

        spsparse_csc = spsp.csc_matrix((2222,22))
        spsparse_csc[1,2] = 44.6

        spsparse_csr = spsp.csr_matrix((2222,22))
        spsparse_csr[1,3] = 44.7

        spsparse_lil = spsp.lil_matrix((2222,22))
        spsparse_lil[3,2] = 44.5

        self.param_dict['Sparse']['lil_mat'] = spsparse_lil
        self.param_dict['Sparse']['csc_mat'] = spsparse_csc
        self.param_dict['Sparse']['csr_mat'] = spsparse_csr

        self.param_dict['Tuples']['int'] = (1,2,3)
        self.param_dict['Tuples']['float'] = (44.4,42.1,3.)
        self.param_dict['Tuples']['str'] = ('1','2wei','dr3i')


    def add_params(self,traj):

        flat_dict = flatten_dictionary(self.param_dict,'.')

        for key, val in flat_dict.items():
            if isinstance(val, (np.ndarray,list, tuple)):
                traj.f_add_parameter(ArrayParameter,key,val, )
            elif isinstance(val, (int,str,bool,float)):
                traj.f_add_parameter(Parameter,key,val, comment='Im a comment!')
            elif spsp.isspmatrix(val):
                traj.f_add_parameter(PickleParameter,key,val).v_annotations.f_set(
                    **{'Name':key,'Val' :str(val),'Favorite_Numbers:':[1,2,3],
                                     'Second_Fav':np.array([43.0,43.0])})
            else:
                raise RuntimeError('You shall not pass, %s is %s!' % (str(val),str(type(val))))

        traj.f_add_derived_parameter('Another.String', 'Hi, how are you?')



    def explore(self, traj):
        self.explored ={'Normal.trial': [0],
            'Numpy.double': [np.array([1.0,2.0,3.0,4.0]), np.array([-1.0,3.0,5.0,7.0])],
            'lil_mat' :[spsp.lil_matrix((2222,22)), spsp.lil_matrix((2222,22))]}

        self.explored['lil_mat'][0][1,2]=44.0
        self.explored['lil_mat'][1][2,2]=33


        traj.f_explore(cartesian_product(self.explored))



    def setUp(self):
        self.set_mode()

        logging.basicConfig(level = logging.DEBUG)

        self.filename = '../../experiments/tests/HDF5/test.hdf5'
        self.logfolder = '../../experiments/tests/Log'
        self.trajname = 'Test'

        env = Environment(trajectory=self.trajname,filename=self.filename,
                          file_title=self.trajname, log_folder=self.logfolder)

        traj = env.v_trajectory

        traj.multiproc = self.multiproc
        traj.wrap_mode = self.mode
        traj.ncores = self.ncores

        traj.v_standard_parameter=Parameter

        ## Create some parameters
        self._create_param_dict()
        ### Add some parameter:
        self.add_params(traj)

        #remember the trajectory and the environment
        self.traj = traj
        self.env = env



    def make_run(self):

        ### Make a test run
        simple_arg = -13
        simple_kwarg= 13.0
        self.env.f_run(simple_calculations,simple_arg,simple_kwarg=simple_kwarg)


    def test_run(self):

        ###Explore
        self.explore(self.traj)

        ###Test, that you cannot append to data
        with self.assertRaises(ValueError):
            self.traj.f_store_item('filename')

        self.make_run()

        newtraj = self.load_trajectory(trajectory_index=-1,as_new=True)
        self.traj.f_update_skeleton()
        self.traj.f_load_items(self.traj.f_to_dict().keys(), only_empties=True)

        self.compare_trajectories(self.traj,newtraj)



    def load_trajectory(self,trajectory_index=None,trajectory_name=None,as_new=True):
        ### Load The Trajectory and check if the values are still the same
        newtraj = Trajectory()
        newtraj.v_storage_service=HDF5StorageService(filename=self.filename)
        newtraj.f_load(trajectory_name=trajectory_name, load_derived_parameters=2,load_results=2,
                       trajectory_index=trajectory_index)
        return newtraj

    def compare_trajectories(self,traj1,traj2):

        old_items = traj1.f_to_dict(fast_access=False)
        new_items = traj2.f_to_dict(fast_access=False)



        self.assertEqual(len(old_items),len(new_items))
        for key,item in new_items.items():
            old_item = old_items[key]
            if key.startswith('config'):
                continue

            if isinstance(item, BaseParameter):
                self.assertTrue(parameters_equal(item,old_item),
                                'For key %s: %s not equal to %s' %(key,str(old_item),str(item)))
            elif isinstance(item,BaseResult):
                self.assertTrue(results_equal(item, old_item),
                                'For key %s: %s not equal to %s' %(key,str(old_item),str(item)))
            else:
                raise RuntimeError('You shall not pass')

            self.assertTrue(nested_equal(item.v_annotations,old_item.v_annotations),'%s != %s' %
                        (item.v_annotations.f_ann_to_str(),old_item.v_annotations.f_ann_to_str()))


class MultiprocQueueTest(EnvironmentTest):

    def set_mode(self):
        self.mode = globally.WRAP_MODE_QUEUE
        self.multiproc = True
        self.ncores = 2


class MultiprocLockTest(EnvironmentTest):

     def set_mode(self):
        self.mode = globally.WRAP_MODE_LOCK
        self.multiproc = True
        self.ncores = 2




if __name__ == '__main__':
    if REMOVE:
        shutil.rmtree('../../Test',True)
    unittest.main()


__author__ = 'Robert Meyer'



import numpy as np
import unittest
from pypet.storageservice import LazyStorageService
from pypet.environment import Environment
import logging


def just_printing_bro(traj):
        key = traj.f_get('Test').gfn()
        value = traj.f_get('par.Test', fast_access=True)
        print 'Current value of %s is %d' %(key, value)

class EnvironmentTest(unittest.TestCase):


    def setUp(self):

        logging.basicConfig(level = logging.DEBUG)

        self.filename = '../../experiments/tests/HDF5/test.hdf5'
        self.logfolder = '../../experiments/tests/Log'
        self.trajname = 'Test'

        env = Environment(self.trajname,self.filename,self.trajname, log_folder=self.logfolder)

        traj = env.v_trajectory
        traj.set_storage_service(LazyStorageService())

        traj.ap('Test', 1)


        large_amount = 111

        for irun in range(large_amount):
            name = 'There.Are.Many.Of.m3' + str(irun)

            traj.ap(name, irun)

        traj.f_preset_config('ncores', 2)
        traj.f_preset_config('multiproc', True)

        traj.f_explore({traj.f_get('par.Test').gfn():[1,2,3,4,5]})

        self.traj = traj

        self.env = env
        self.traj = traj

    def test_multiprocessing(self):

        self.env.f_run(just_printing_bro)


if __name__ == '__main__':
    unittest.main()



import logging


class StreamToLogger(object):
    """
    Fake file-like stream object that redirects writes to a logger instance.
    """
    def __init__(self, logger, log_level=logging.INFO):
        self._logger = logger
        self._log_level = log_level
        self._linebuf = ''
 
    def write(self, buf):
        for line in buf.rstrip().splitlines():
            self._logger.log(self._log_level, line.rstrip())
    
    def flush(self):
        pass










'''
Created on 17.05.2013
'''

class ParameterLockedException(TypeError):
    '''Exception raised if someone tries to modify a locked ParameterSet'''
    def __init__(self,msg):
        self._msg=msg
        
    def __str__(self):
        return repr(self._msg)
    
class ParameterNotArrayException(TypeError):
    '''Exception raised if someone tries to treat the Parameter as an Parameter Array if it is not'''
    def __init__(self,msg):
        self._msg=msg
        
    def __str__(self):
        return repr(self._msg)


class DefaultReplacementError(Exception):
    '''Exception raised if added parameters should actually replace default settings, but have not probably due to a type.'''
    def __init__(self,msg):
        self._msg=msg

    def __str__(self):
        return repr(self._msg)
    
class NoSuchServiceError(TypeError):
    '''Exception raised by the Storage Service if a specific operation is not supported, i.e. the message is not understood.
    '''
    def __init__(self,msg):
        self._msg=msg

    def __str__(self):
        return repr(self._msg)


class NotUniqueNodeError(AttributeError):
    '''Exception raised by the Natural Naming if a node can be found more than once.
    '''
    def __init__(self,msg):
        self._msg=msg

    def __str__(self):
        return repr(self._msg)

class TooManyGroupsError(TypeError):
    '''Exception raised by natural naming fast search if fast search cannot be applied
    '''
    def __init__(self,msg):
        self._msg=msg

    def __str__(self):
        return repr(self._msg)


__author__ = 'Robert Meyer'


from collections import Sequence, Mapping, Set
import numpy as np
import pandas as pd
from pypet import globally



def copydoc(fromfunc, sep="\n"):
    """
    Decorator: Copy the docstring of `fromfunc`
    """
    def _decorator(func):
        sourcedoc = fromfunc.__doc__
        if func.__doc__ == None:
            func.__doc__ = sourcedoc
        else:
            func.__doc__ = sep.join([sourcedoc, func.__doc__])
        return func
    return _decorator



def flatten_dictionary(nested_dict, separator):
    flat_dict = {}
    for key, val in nested_dict.items():
        if isinstance(val,dict):
            new_flat_dict = flatten_dictionary(val, separator)
            for flat_key, val in new_flat_dict.items():
                new_key = key + separator + flat_key
                flat_dict[new_key] = val
        else:
            flat_dict[key] = val

    return flat_dict

def nest_dictionary(flat_dict, separator):
    nested_dict = {}
    for key, val in flat_dict.items():
        split_key = key.split(separator)
        act_dict = nested_dict
        final_key = split_key.pop()
        for new_key in split_key:
            if not new_key in act_dict:
                act_dict[new_key] ={}

            act_dict = act_dict[new_key]

        act_dict[final_key] = val
    return nested_dict




def nested_equal(a, b):
    """
    Compare two objects recursively by element, handling numpy objects.

    Assumes hashable items are not mutable in a way that affects equality.
    """

    # for types that implement their own custom strict equality checking

    if hasattr(a,'__eq__'):
        try:
            custom_eq= a == b
            if isinstance(custom_eq,bool):
                return custom_eq
        except ValueError:
            pass



    #Check equality according to type type [sic].
    if a is None:
        return b is None
    if isinstance(a, basestring):
         return a == b
    if isinstance(a, globally.PARAMETER_SUPPORTED_DATA):
        return a==b
    if isinstance(a, np.ndarray):
        return np.all(a==b)
    if isinstance(a, pd.DataFrame):
        new_frame = a == b
        new_frame = new_frame |( pd.isnull(a) & pd.isnull(b))
        return np.all(new_frame)
    if isinstance(a, Sequence):
        return all(nested_equal(x, y) for x, y in zip(a, b))
    if isinstance(a, Mapping):
        if set(a.keys()) != set(b.keys()):
            return False
        return all(nested_equal(a[k], b[k]) for k in a.keys())
    if isinstance(a, Set):
         return a == b

    if hasattr(a,'__dict__'):
        if not hasattr(b,'__dict__'):
            return False

        if set(a.__dict__.keys()) != set(b.__dict__.keys()):
            return False

        return all(nested_equal(a.__dict__[k], b.__dict__[k]) for k in a.__dict__.keys())

    return id(a) == id(b)








__author__ = 'Robert Meyer'

import numpy as np
from pypet.parameter import Parameter
from pypet.trajectory import Trajectory, SingleRun
from pypet.storageservice import LazyStorageService
import pickle
import logging
import cProfile

def prepare():
    logging.basicConfig(level = logging.INFO)
    traj = Trajectory('Test')



    traj.set_storage_service(LazyStorageService())


    large_amount = 111111

    for irun in range(large_amount):
        name = 'Hurz.L' + str(irun)

        traj.ap(name,irun)


    traj.ap('Hurz.Test', value=1)

    return traj


def print_val():
    print traj.f_get(what)

def main():
    global traj
    global what
    traj=prepare()


    what = 'Test'

    cProfile.run('print_val()',sort=1)

if __name__ == '__main__':
    main()


__version__ = "unknown"
try:
    from _version import __version__
except ImportError:
    # We're running in a tree that doesn't
    # have a _version.py, so we don't know what our version is.
    pass




__author__ = 'Robert Meyer'



import numpy as np
import unittest
from pypet.parameter import Parameter, PickleParameter, Result, BaseResult
from pypet.trajectory import Trajectory, SingleRun
from pypet.storageservice import LazyStorageService
import pickle
import logging
import cProfile
import scipy.sparse as spsp
import pypet.petexceptions as pex
import multiprocessing as multip

import pypet.storageservice as stsv

class ImAParameterInDisguise(Parameter):
    pass

class ImAResultInDisguise(Result):
    pass

class TrajectoryTest(unittest.TestCase):


    def setUp(self):
        name = 'Moop'

        self.traj = Trajectory(name,[ImAParameterInDisguise])

        comment = 'This is a comment'
        self.traj.v_comment=comment

        self.assertTrue(comment == self.traj.v_comment)

        self.traj.f_add_parameter('IntParam',3)
        sparsemat = spsp.csr_matrix((1000,1000))
        sparsemat[1,2] = 17.777

        self.traj.f_add_parameter(PickleParameter,'SparseParam', sparsemat)

        self.traj.f_add_parameter('FloatParam')

        self.traj.f_add_derived_parameter(Parameter('FortyTwo', 42))

        self.traj.f_add_result(Result,'Im.A.Simple.Result',44444)

        self.traj.FloatParam=4.0

        self.traj.f_explore({'FloatParam':[1.0,1.1,1.2,1.3]})

        self.assertTrue(len(self.traj) == 4)

        self.traj.f_add_parameter_group('peter.paul')

        self.traj.f_add_parameter('peter.markus.yve',6)

        self.traj.peter.f_add_parameter('paul.peter')

        with self.assertRaises(AttributeError):
            self.traj.markus.peter



        with self.assertRaises(AttributeError):
            self.traj.f_add_parameter('Peter.  h ._hurz')



    def testGet(self):
        self.traj.v_fast_access=True
        self.traj.f_get('FloatParam', fast_access=True) == self.traj.FloatParam

        self.traj.v_fast_access=False

        self.assertTrue(self.traj.f_get('FloatParam').f_get() == 4.0 , '%d != 4.0' %self.traj.f_get('FloatParam').f_get())

        self.assertTrue(self.traj.FortyTwo.f_get() == 42)


    def testremove(self):
        self.traj.f_remove_item(self.traj.f_get('peter.markus.yve'),remove_empty_groups=True)

        with self.assertRaises(AttributeError):
            self.traj.peter.markus.yve

        self.assertFalse('peter.markus.yve' in self.traj)

        #self.assertTrue(len(self.traj)==1)

        self.traj.f_remove_item('FortyTwo',remove_empyt_groups=True)

        self.traj.f_remove_item('SparseParam')
        self.traj.f_remove_item('IntParam')

        #self.assertTrue(len(self.traj)==1)

    def test_changing(self):

        self.traj.f_preset_config('testconf', 1)
        self.traj.f_preset_parameter('testparam', 1)
        self.traj.f_preset_parameter('I_do_not_exist', 2)

        self.traj.f_add_parameter('testparam', 0)
        self.traj.f_add_config('testconf', 0)

        self.traj.v_fast_access=True

        self.assertTrue(self.traj.testparam == 1)
        self.assertTrue(self.traj.testconf == 1)

        ### should raise an error because 'I_do_not_exist', does not exist:
        with self.assertRaises(pex.DefaultReplacementError):
            self.traj._prepare_experiment()


    def test_if_pickable(self):

        self.traj.v_fast_access=True

        #self.traj.v_full_copy=True
        dump = pickle.dumps(self.traj)

        newtraj = pickle.loads(dump)


        self.assertTrue(len(newtraj) == len(self.traj))

        new_items = newtraj.f_to_dict(fast_access=True)

        for key, val in self.traj.f_to_dict(fast_access=True).items():
            #val = newtraj.f_get(key)
            nval = new_items[key]
            if isinstance(val, BaseResult):
                for ikey in val._data:
                    self.assertTrue(str(nval.f_get(ikey))==str(val.f_get(ikey)))
            else:

                self.assertTrue(str(val)==str(nval), '%s != %s' %(str(val),str(nval)))

    def test_dynamic_class_loading(self):
        self.traj.f_add_parameter(ImAParameterInDisguise,'Rolf', 1.8)

    def test_standard_change_param_change(self):
        self.traj.v_standard_parameter=ImAParameterInDisguise

        self.traj.f_add_parameter('I.should_be_not.normal')

        self.assertIsInstance(self.traj.f_get('normal'), ImAParameterInDisguise,'Param is %s insted of ParamInDisguise.' %str(type(self.traj.normal)))

        self.traj.v_standard_result=ImAResultInDisguise

        self.traj.f_add_result('Peter.Parker')

        self.assertIsInstance(self.traj.Parker, ImAResultInDisguise)

class TrajectoryMergeTest(unittest.TestCase):

    def setUp(self):
        name = 'Moop'

        self.traj = Trajectory(name,[ImAParameterInDisguise])

        comment = 'This is a comment'
        self.traj.v_comment=comment

        self.assertTrue(comment == self.traj.v_comment)

        self.traj.f_add_parameter('IntParam',3)
        sparsemat = spsp.csr_matrix((1000,1000))
        sparsemat[1,2] = 17.777

        #self.traj.f_add_parameter('SparseParam', sparsemat, param_type=PickleParameter)

        self.traj.f_add_parameter('FloatParam')

        self.traj.f_add_derived_parameter(Parameter('FortyTwo', 42))
        self.traj.f_add_parameter('Trials',0)

        self.traj.f_add_result(Result,'Im.A.Simple.Result',44444)

        self.traj.FloatParam=4.0
        self.traj.v_storage_service = LazyStorageService()


        self.traj.f_explore({'FloatParam':[1.0,1.1,1.2,1.3],'Trials':[0,1,2,3]})


        self.assertTrue(len(self.traj) == 4)


        name2 = 'aaaaah'
        self.traj2 = Trajectory(name2,[ImAParameterInDisguise])

        comment = 'This is a comment'
        self.traj2.v_comment=comment

        self.assertTrue(comment == self.traj2.v_comment)

        self.traj2.f_add_parameter('IntParam',3)
        sparsemat = spsp.csr_matrix((1000,1000))
        sparsemat[1,2] = 17.777

        #self.traj2.f_add_parameter('SparseParam', sparsemat, param_type=PickleParameter)
        self.traj2.f_add_parameter('Trials',0)

        self.traj2.f_add_parameter('FloatParam')

        self.traj2.f_add_derived_parameter(Parameter('FortyTwo', 42))

        self.traj2.f_add_result(Result,'Im.A.Simple.Result',44444)

        self.traj2.FloatParam=4.0

        self.traj2.f_explore({'FloatParam':[42.0,43.0,1.2,1.3],'Trials':[0,1,2,3]})
        self.traj2.v_storage_service = LazyStorageService()

        self.assertTrue(len(self.traj2) == 4)



    def test_merge_parameters_without_remove(self):
        # remove_duplicates = True should be discarded by the trial parameter
        self.traj._merge_parameters(self.traj2,trial_parameter='Trials',remove_duplicates=True)

    def test_merge_parameters_with_remove(self):
        self.traj._merge_parameters(self.traj2,remove_duplicates=True)

    def test_merge_without_remove(self):
        self.traj.f_merge(self.traj2, remove_duplicates=True,trial_parameter='Trials')

    def test_merge_with_remove(self):
        self.traj.f_merge(self.traj2, remove_duplicates=True)

class SingleRunTest(unittest.TestCase):


    def setUp(self):

        logging.basicConfig(level = logging.INFO)
        traj = Trajectory('Test')

        traj.v_storage_service = LazyStorageService()

        large_amount = 111

        for irun in range(large_amount):
            name = 'There.Are.Many.Parameters.Like.Me' + str(irun)

            traj.f_add_parameter(name, irun)


        traj.f_add_parameter('TestExplorer', 1)

        traj.v_fast_access=False
        traj.f_explore({traj.TestExplorer.v_full_name:[1,2,3,4,5]})
        traj.v_fast_access=True

        self.traj = traj
        self.n = 1
        self.single_run = self.traj._make_single_run(self.n)
        self.assertTrue(len(self.single_run)==1)



    def test_if_single_run_can_be_pickled(self):

        self.single_run._storageservice=stsv.QueueStorageServiceSender()
        dump = pickle.dumps(self.single_run)

        single_run_rec = pickle.loads(dump)

        #print single_run.f_get('Test').val

        elements_dict = self.single_run.f_to_dict()
        for key in elements_dict:
            val = self.single_run.f_get(key,fast_access=True)
            val_rec = single_run_rec.f_get(key).f_get()
            self.assertTrue(np.all(val==val_rec))



    def test_adding_derived_parameter_and_result(self):
        value = 44.444
        self.single_run.f_add_derived_parameter('Im.A.Nice.Guy.Yo', value)
        self.assertTrue(self.single_run.Nice.Yo == value)

        self.single_run.f_add_result('Puberty.Hacks', val=value)
        resval= self.single_run.Hacks.f_get('val')
        self.assertTrue(resval == value, '%s != %s' % ( str(resval),str(value)))


    def test_standard_change_param_change(self):
        self.single_run.v_standard_parameter=ImAParameterInDisguise

        self.single_run.f_add_derived_parameter('I.should_be_not.normal')

        self.assertIsInstance(self.single_run.f_get('normal'), ImAParameterInDisguise)

        self.single_run.v_standard_result=ImAResultInDisguise

        self.single_run.f_add_result('Peter.Parker')

        self.assertIsInstance(self.single_run.Parker, ImAResultInDisguise)

class SingleRunQueueTest(unittest.TestCase):


    def setUp(self):

        logging.basicConfig(level = logging.INFO)
        traj = Trajectory('Test')

        traj.v_storage_service=LazyStorageService()

        large_amount = 11

        for irun in range(large_amount):
            name = 'There.Are.Many.Parameters.Like.Me' + str(irun)

            traj.f_add_parameter(name, irun)


        traj.f_add_parameter('TestExplorer', 1)

        traj.f_explore({traj.f_get('TestExplorer').v_full_name:[1,2,3,4,5]})

        self.traj = traj
        self.n = 1
        self.single_run = self.traj._make_single_run(self.n)
        self.assertTrue(len(self.single_run)==1)


    def test_queue(self):

        manager = multip.Manager()
        queue = manager.Queue()

        to_put = ('msg',[self.single_run],{})

        queue.put(to_put)

        pass


if __name__ == '__main__':
    unittest.main()